#!/usr/bin/perl -w 

# Copyright 2010-2012 Kevin Gimpel
# Licensed under the modified BSD license. See the LICENSE file included with this distribution.

# Usage:
# run-rampion-reranker.pl <source_file> <target_file> <decoder-executable> <decoder-config>
# For other options see below or run 'run-rampion-reranker.pl --help'
#
# Notes:
# <source_file> and <target_file> should be raw text files, one sentence per line
# <target_file> can be a prefix, in which case the files are <target_file>0, <target_file>1, etc. are used
#
# This file was created by modifying run-rampion-moses.pl.
#
# Since this file contains some code from mert-moses.pl, which is released under an LGPL
# license, this file is also released under an LGPL license. See the file LICENSE in the same 
# directory as this file.

my $verbose = 0;
my $usage = 0; # request for --help
my $working_dir = "rampion-working";
my $tune_src = undef; # required, input text to decode
my $tune_ref = undef; # required, basename of files with references
my $config = undef; # required, pathname to startup ini file

my $rampion_cmd = "./rampion";
my $loss = "ramp3";
my $num_iters = "20";
my $num_cccpiters = "10";
my $num_subiters = "5";
my $opt_alg = "ssd";
my $eta = "0.0001";
my $C = "1.0";
my $alpha = "10.0";
my $fastmode = 0;
my $cost_metric = "LIN_OCH_BLEU_PLUS_ONE";
my $nbest_file = "DUMMY";

use strict;
use Getopt::Long;
GetOptions(
  "working-dir=s" => \$working_dir,
  "input=s" => \$tune_src,
  "refs=s" => \$tune_ref,
  "config=s" => \$config,
  "help" => \$usage,
  "verbose" => \$verbose,
  "rampion-cmd=s" => \$rampion_cmd,
  "loss=s" => \$loss,
  "num-iters=i" => \$num_iters,
  "num-cccpiters=s" => \$num_cccpiters,
  "num-subiters=s" => \$num_subiters,
  "opt-alg=s" => \$opt_alg,
  "eta=s" => \$eta,
  "C=s" => \$C,
  "alpha=s" => \$alpha,
  "fast-mode" => \$fastmode,
  "cost-metric=s" => \$cost_metric,
  "nbest-list-file=s" => \$nbest_file
) or exit(1);

if (scalar @ARGV == 3) {
  # required parameters: input_file references_basename decoder_config_file
  $tune_src = shift;
  $tune_ref = shift;
  $config = shift;
}

if ($usage || !defined $tune_src || !defined $tune_ref || !defined $config) {
  print STDERR "usage: run-rampion-reranker.pl input-text references decoder.ini
Options:
  --working-dir=rampion-dir ... where all output files are generated
  --rampion-cmd=STRING   ... command to call rampion (absolute path and executable)
  --loss=STRING  ... rampion loss function (default: $loss)
  --num-iters=ITERS ... number of training iterations (default: $num_iters)
  --num-cccpiters=INT ... number of iterations of CCCP to run per larger iteration (default: $num_cccpiters)
  --num-subiters=INT ... number of iterations of optimization to run per CCCP iteration (default: $num_subiters)
  --opt-alg=STRING ... optimization algorithm to use within Rampion; only relevant for certain loss functions (default: $opt_alg)
  --eta=FLOAT  ... fixed stepsize for optimization algorithm (default: $eta)
  --C=FLOAT  ... L2 regularization coefficient (default: $C)
  --alpha=FLOAT ... coefficient for cost term (default: $alpha)
  --fast-mode ... if present, will run in Rampion in \"fast mode\", computing fewer statistics; useful with large tuning sets (default: not fast mode)
  --cost-metric=STRING ... MT evaluation metric to use in cost function. choices: STANDARD_BLEU, LIN_OCH_BLEU_PLUS_ONE, BLEU_PLUS_ONE_ALL_N, BLEU_PLUS_ONE_ALL_N_PGROUND, BLEU_PLUS_ONE_ALL_N_AND_BP, BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND (default: LIN_OCH_BLEU_PLUS_ONE)
  --nbest-list-file=STRING ... file containing n-best lists with additional features
";
  exit 1;
}

my $input_abs = ensure_full_path($tune_src);
die "File not found: $tune_src (interpreted as $input_abs)."
  if ! -e $input_abs;
$tune_src = $input_abs;

my $ref_abs = ensure_full_path($tune_ref);
# Check if references exist and store a list of them.
my @references;
if (-e $ref_abs) {
  push @references, $ref_abs;
} else {
  # If there are multiple files, get a full list of them.
  my $part = 0;
  while (-e $ref_abs.$part) {
    push @references, $ref_abs.$part;
    $part++;
  }
  die("Reference translations not found: $tune_ref (interpreted as $ref_abs)") unless $part;
}

my $config_abs = ensure_full_path($config);
die "File not found: $config (interpreted as $config_abs)."
  if ! -e $config_abs;
$config = $config_abs;

my $nbest_file_abs = ensure_full_path($nbest_file);
die "File not found: $nbest_file (interpreted as $nbest_file_abs)."
  if ! -e $nbest_file_abs;
$nbest_file = $nbest_file_abs;

# Store current directory and create the working directory if needed.
my $cwd = `pawd 2>/dev/null`; 
if(!$cwd){$cwd = `pwd`;}
chomp($cwd);
safesystem("mkdir -p $working_dir") or die "Can't mkdir $working_dir";

{
# Open local scope.

# Change to the working directory.
chdir($working_dir) or die "Can't chdir to $working_dir";

# Copy the initial moses.ini file to where the decoder will read it on the first iteration.
safesystem("\\cp -f $config ./run1.moses.ini") or die;

my $rampion_fast_mode = "false";
if ($fastmode == 1) {
  $rampion_fast_mode = "true";
}

for (my $run = 1; $run <= $num_iters; $run++) {
  # Run decoder to produce n-best lists.
  print "Iteration $run started at ".`date`;

  if ($run == 1) {
    # If we're on the first iteration, create an empty file for the previous k-best lists.
    safesystem("touch prev_nbest_file.uniq") or die "Failed to create empty prev_nbest_file.uniq";
  }

  # At this point, the current n-best lists should be in $nbest_file and the 
  # previous k-best lists should be in prev_nbest_file.uniq.

  # Increment the iteration counter.
  my $nextrun = $run + 1;

  # Call rampion, passing the n-best list file and all other parameters.
  my $cmd = "$rampion_cmd src_file=$tune_src kbest_file=$nbest_file ".
	"prev_kbest_file=prev_nbest_file.uniq ref_file=$ref_abs ".
	"in_moses_ini_file=./run$run.moses.ini out_moses_ini_file=./run$nextrun.moses.ini ".
	"initial_moses_ini_file=$config loss_function=$loss ".
	"curr_train_iter=$nextrun num_cccpiters=$num_cccpiters num_subiters=$num_subiters ".
	"optimization_algorithm=$opt_alg eta=$eta C=$C alpha=$alpha fast_mode=$rampion_fast_mode cost_function_metric=$cost_metric";
  safesystem("$cmd > rampion.iter$run.out 2> rampion.iter$run.err") or die "Failed to run rampion.";
}

print "Training finished at ".`date`;

#chdir back to the original directory # useless, just to remind we were not there
chdir($cwd);

} # end of local scope

sub safesystem {
  print STDERR "Executing: @_\n";
  system(@_);
  if ($? == -1) {
      print STDERR "Failed to execute: @_\n  $!\n";
      exit(1);
  }
  elsif ($? & 127) {
      printf STDERR "Execution of: @_\n  died with signal %d, %s coredump\n",
          ($? & 127),  ($? & 128) ? 'with' : 'without';
      exit(1);
  }
  else {
    my $exitcode = $? >> 8;
    print STDERR "Exit code: $exitcode\n" if $exitcode;
    return ! $exitcode;
  }
}

sub ensure_full_path {
    my $PATH = shift;
$PATH =~ s/\/nfsmnt//;
    return $PATH if $PATH =~ /^\//;
    my $dir = `pawd 2>/dev/null`; 
    if(!$dir){$dir = `pwd`;}
    chomp($dir);
    $PATH = $dir."/".$PATH;
    $PATH =~ s/[\r\n]//g;
    $PATH =~ s/\/\.\//\//g;
    $PATH =~ s/\/+/\//g;
    my $sanity = 0;
    while($PATH =~ /\/\.\.\// && $sanity++<10) {
        $PATH =~ s/\/+/\//g;
        $PATH =~ s/\/[^\/]+\/\.\.\//\//g;
    }
    $PATH =~ s/\/[^\/]+\/\.\.$//;
    $PATH =~ s/\/+$//;
$PATH =~ s/\/nfsmnt//;
    return $PATH;
}
