// Copyright 2010-2012 Kevin Gimpel
// Licensed under the modified BSD license. See the LICENSE file included with this distribution.
#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <math.h>
#include "mteval.h"

// multiReferenceLines should hold one vector for each reference file, and for each, a vector
// holding that reference file's lines.
MTEval::MTEval(const vector< vector<string> >& multiReferenceLines, Metric metric) {
	initCommon(metric);
	m_numReferences = multiReferenceLines.size();
	updateMultiReferenceLines(multiReferenceLines);
}

MTEval::~MTEval() {}

void MTEval::initCommon(Metric metric) {
	m_metric = metric;
	m_ngramOrder = 4;
	m_translationLength = 0;
	m_referenceLength = 0;
	m_numReferences = 1;
	resetBLEUStats();
}

void MTEval::updateReference(const string& reference) {
	vector<int> referenceLengths;
	int referenceLength;

	map<string, int> referenceLineNGrams;
	fillReferenceNGrams(reference, referenceLineNGrams, &referenceLength);
	m_referenceLineNGrams.push_back(referenceLineNGrams);

	referenceLengths.push_back(referenceLength);
	m_multiReferenceLengths.push_back(referenceLengths);
}

void MTEval::updateReferenceLines(const vector<string>& referenceLines) {
	m_referenceLength = 0;
	for (unsigned int i = 0; i < referenceLines.size(); ++i) {
		map<string, int> referenceLineNGrams;
		int referenceLength;
		fillReferenceNGrams(referenceLines[i], referenceLineNGrams, &referenceLength);
		m_referenceLineNGrams.push_back(referenceLineNGrams);

		vector<int> referenceLengths;
		referenceLengths.push_back(referenceLength);
		m_multiReferenceLengths.push_back(referenceLengths);

		m_referenceLength += referenceLength;
	}
}

// multiReferenceLines holds one vector for each reference file, and for each, a vector
// holding that reference file's lines.
void MTEval::updateMultiReferenceLines(const vector< vector<string> >& multiReferenceLines) {
	m_referenceLength = 0;

	m_multiReferenceLengths.clear();
	m_referenceLineNGrams.clear();
	int numLines = 0;
	if (multiReferenceLines.size() > 0) {
		numLines = multiReferenceLines[0].size();
	}
	// Go through all lines.
	for (unsigned int j = 0; j < numLines; ++j) {
		// For each line, go through all reference translations for the line.
		map<string, int> multiReferenceLineNGrams;
		vector<int> multiReferenceLengths;
		for (unsigned int i = 0; i < multiReferenceLines.size(); ++i) {
			int refLength;

			map<string, int> referenceLineNGrams;
			fillReferenceNGrams(multiReferenceLines[i][j], referenceLineNGrams, &refLength);
			// Store the reference length.
			multiReferenceLengths.push_back(refLength);
			// Merge the n-grams in this reference sentence with the n-grams in preceding
			// reference sentences, while only keeping the max-count that any particular n-gram
			// is seen with (per BLEU's policy).
			mergeReferenceNGrams(multiReferenceLineNGrams, referenceLineNGrams);
		}

		m_multiReferenceLengths.push_back(multiReferenceLengths);
		m_referenceLineNGrams.push_back(multiReferenceLineNGrams);
	}
}

void MTEval::addReferenceLine(const string& referenceLine) {
	map<string, int> referenceLineNGrams;
	int refLength;
	fillReferenceNGrams(referenceLine, referenceLineNGrams, &refLength);
	m_referenceLineNGrams.push_back(referenceLineNGrams);
	if (m_multiReferenceLengths.size() == 0) {
		vector<int> refLengths;
		refLengths.push_back(refLength);
		m_multiReferenceLengths.push_back(refLengths);
	} else {
		m_multiReferenceLengths[0].push_back(refLength);
	}
	m_referenceLength += refLength;
}

void MTEval::resetBLEUStats() {
	m_ngramMatches.clear();
	m_ngramTotals.clear();
	m_ngramMatches.resize(m_ngramOrder, 0);
	m_ngramTotals.resize(m_ngramOrder, 0);
	// If we're using a smoothed version of BLEU, add 1 to matches and totals for certain n-grams.
	if (m_metric == LIN_OCH_BLEU_PLUS_ONE || m_metric == BLEU_PLUS_ONE_ALL_N || m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND 
		|| m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {

		if (m_metric == BLEU_PLUS_ONE_ALL_N || m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND || 
			m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {

			m_ngramMatches[0] = 1;
			m_ngramTotals[0] = 1;
		}
		for (int i = 1; i < m_ngramOrder; ++i) {
			m_ngramMatches[i] = 1;
			m_ngramTotals[i] = 1;
		}
	}
	m_translationLength = 0;
	m_referenceLength = 0;
}

void MTEval::addBLEUStats(const BLEUSuffStats& bleuStats) {
	for (unsigned int i = 0; i < bleuStats.ngramMatches.size(); ++i) {
		m_ngramMatches[i] += bleuStats.ngramMatches[i];
	}
	for (unsigned int i = 0; i < bleuStats.ngramTotals.size(); ++i) {
		m_ngramTotals[i] += bleuStats.ngramTotals[i];
	}
	m_translationLength += bleuStats.translationLength;
	m_referenceLength += bleuStats.referenceLength;
}

void MTEval::subtractBLEUStats(const BLEUSuffStats& bleuStats) {
	for (unsigned int i = 0; i < bleuStats.ngramMatches.size(); ++i) {
		m_ngramMatches[i] -= bleuStats.ngramMatches[i];
	}
	for (unsigned int i = 0; i < bleuStats.ngramTotals.size(); ++i) {
		m_ngramTotals[i] -= bleuStats.ngramTotals[i];
	}
	m_translationLength -= bleuStats.translationLength;
	m_referenceLength -= bleuStats.referenceLength;
}

// This function assumes that m_ngramMatches, m_ngramTotals, m_translationLength, and m_referenceLength contain
// sufficient statistics for computing BLEU.
double MTEval::computeBLEU() {
	double bleu = 0.0;
	vector<double> ngramPrecisions(m_ngramOrder, 0.0);

	for (int i = 0; i < m_ngramOrder; ++i) {
		if (m_ngramTotals[i] > 0) {
			ngramPrecisions[i] = ((double) m_ngramMatches[i]) / ((double) m_ngramTotals[i]);
		} else {
			ngramPrecisions[i] = 0.0;
		}
	}

	bleu = MTEval::computeBrevityPenalty(m_translationLength, m_referenceLength);

	double precTerm = 0.0;
	for (int i = 0; i < m_ngramOrder; ++i) {
		if (ngramPrecisions[i] == 0.0) {
			// If any of the n-gram precisions is 0.0, the BLEU score is 0.0.
			bleu = 0.0;
			break;
		} else {
			precTerm += log(ngramPrecisions[i]);
		}
	}
	bleu *= exp(precTerm / 4);

	return bleu;
}

// This function assumes that m_multiReferenceLengths and m_referenceLineNGrams have already been filled in.
double MTEval::computeBLEU(const vector<string>& translationLines) {
	string printScore;
	double bleu = computeBLEUString(translationLines, printScore);
	//printf("BLEU = %.2f, %.1f/%.1f/%.1f/%.1f (BP=%.3f, ratio=%.3f, hyp_len=%d, ref_len=%d)\n",
	//	100*bleu, 100*ngramPrecisions[0], 100*ngramPrecisions[1], 100*ngramPrecisions[2], 100*ngramPrecisions[3],
	//	brevityPenalty, (((double) cumulativeTranslationLength) / ((double) cumulativeReferenceLength)),
	//	cumulativeTranslationLength, cumulativeReferenceLength);
	return bleu;
}

double MTEval::computeBLEUString(const vector<string>& translationLines, string& printScore) {
	if (m_ngramOrder <= 0) {
		return 0.0;
	}

	// Initialize counts of n-gram matches and total number of n-grams for each n-gram length.
	vector<int> ngramMatches(m_ngramOrder, 0);
	vector<int> ngramTotals(m_ngramOrder, 0);
	if (m_metric == LIN_OCH_BLEU_PLUS_ONE || m_metric == BLEU_PLUS_ONE_ALL_N || m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND 
		|| m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {

		if (m_metric == BLEU_PLUS_ONE_ALL_N || m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND
			|| m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {

			ngramMatches[0] = 1;
			ngramTotals[0] = 1;
		}
		for (int i = 1; i < m_ngramOrder; ++i) {
			ngramMatches[i] = 1;
			ngramTotals[i] = 1;
		}
	}

	int cumulativeTranslationLength = 0;
	int cumulativeReferenceLength = 0;

	for (unsigned int j = 0; j < translationLines.size(); ++j) {
		vector<string> tokens;
		// Tokenize the translation.
		tokenize(translationLines[j], ' ', tokens);
		int translationLength = tokens.size();
		map<string, int> translationNGrams;
		fillNGramCountsFromTokens(tokens, translationNGrams);

		// Now go back through all n-grams in the translation and get counts of matches.
		map<string, int>::const_iterator ngramIter = translationNGrams.begin();
		map<string, int>::const_iterator refIter;
		for (; ngramIter != translationNGrams.end(); ++ngramIter) {
			// Get the count of the n-gram in the translation.
			int transCount = ngramIter->second;
			// Get the number of tokens in the n-gram.
			int ngramLength = countChar(ngramIter->first, ' ') + 1;
			if (ngramLength <= 0 || ngramLength > m_ngramOrder) {
				cerr << "Error: invalid n-gram length: " << ngramLength << endl;
			} else {
				ngramTotals[ngramLength-1] += transCount;

				// Search for the current n-gram in the reference.
				refIter = m_referenceLineNGrams[j].find(ngramIter->first);
				if (refIter != m_referenceLineNGrams[j].end()) {
					int refCount = refIter->second;
					if (refCount >= transCount) {
						ngramMatches[ngramLength-1] += transCount;
					} else {
						// There are more occurrences in the translation than in the reference,
						// so we only get credit for the count in the reference.
						ngramMatches[ngramLength-1] += refCount;
					}
				}
			}
		}

		cumulativeTranslationLength += translationLength;
		if (m_numReferences == 1) {
			cumulativeReferenceLength += m_multiReferenceLengths[j][0];
		} else if (m_numReferences > 1) {
			// Find the reference sentence with length closest to the current translation.
			int minAbsLenDiff = 99999;
			int closestRefLen = -1;
			for (unsigned int i = 0; i < m_multiReferenceLengths[j].size(); ++i) {
				int absLenDiff = abs(translationLength - m_multiReferenceLengths[j][i]);
				bool makeUpdate = false;
				if (absLenDiff < minAbsLenDiff) {
					makeUpdate = true;
				}
				if (absLenDiff == minAbsLenDiff && m_multiReferenceLengths[j][i] < closestRefLen) {
					makeUpdate = true;
				}

				if (makeUpdate) {
					minAbsLenDiff = absLenDiff;
					closestRefLen = m_multiReferenceLengths[j][i];
				}
			}
			if (closestRefLen == -1) {
				cerr << "Error: while computing BLEU with multiple references, closest reference length = -1!" << endl;
			} else {
				cumulativeReferenceLength += closestRefLen;
			}
		} else {
			cerr << "Error: while computing BLEU, 0 references given!" << endl;
		}
	}

	vector<double> ngramPrecisions(m_ngramOrder, 0.0);
	for (int i = 0; i < m_ngramOrder; ++i) {
		if (ngramTotals[i] > 0) {
			ngramPrecisions[i] = ((double) ngramMatches[i]) / ((double) ngramTotals[i]);
		} else {
			ngramPrecisions[i] = 0.0;
		}
	}

	// Compute the brevity penalty.
	double brevityPenalty = MTEval::computeBrevityPenalty(cumulativeTranslationLength, cumulativeReferenceLength);

	double bleu = brevityPenalty;
	double precTerm = 0.0;
	for (int i = 0; i < m_ngramOrder; ++i) {
		if (ngramPrecisions[i] == 0.0) {
			// If any of the n-gram precisions is 0.0, the BLEU score is 0.0.
			bleu = 0.0;
			break;
		} else {
			precTerm += log(ngramPrecisions[i]);
		}
	}
	bleu *= exp(precTerm / 4);

	char scoreStringFormat[2000];
	sprintf(scoreStringFormat, "BLEU = %.2f, %.1f/%.1f/%.1f/%.1f (BP=%.3f, ratio=%.3f, hyp_len=%d, ref_len=%d)",
		100*bleu, 100*ngramPrecisions[0], 100*ngramPrecisions[1], 100*ngramPrecisions[2], 100*ngramPrecisions[3],
		brevityPenalty, (((double) cumulativeTranslationLength) / ((double) cumulativeReferenceLength)),
		cumulativeTranslationLength, cumulativeReferenceLength);
	printScore = scoreStringFormat;
	return bleu;
}

// Computes BLEU of the given sentence (translation), placing the computed BLEU sufficient statistics into bleuStats. 
// This function assumes that all reference translations for the given sentence have already been added to 
// this MTEval object as the first reference line (i.e., m_multiReferenceLengths and m_referenceLineNGrams 
// must be filled in and entry 0 is used in each).
double MTEval::computeSentenceBLEU(const string& translation, BLEUSuffStats& bleuStats) {

	if (m_ngramOrder > 0) {
		vector<string> tokens;
		// Tokenize the translation.
		tokenize(translation, ' ', tokens);
		int translationLength = tokens.size();
		map<string, int> translationNGrams;
		fillNGramCountsFromTokens(tokens, translationNGrams);

		// Initialize counts of n-gram matches and total number of n-grams for each n-gram length.
		bleuStats.ngramMatches.clear();
		bleuStats.ngramTotals.clear();
		bleuStats.ngramMatches.resize(m_ngramOrder, 0);
		bleuStats.ngramTotals.resize(m_ngramOrder, 0);

		if (m_metric == LIN_OCH_BLEU_PLUS_ONE || m_metric == BLEU_PLUS_ONE_ALL_N || m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND
			|| m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {

			if (m_metric == BLEU_PLUS_ONE_ALL_N || m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND 
				|| m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {

				bleuStats.ngramMatches[0] = 1;
				bleuStats.ngramTotals[0] = 1;
			}

			for (int i = 1; i < m_ngramOrder; ++i) {
				bleuStats.ngramMatches[i] = 1;
				bleuStats.ngramTotals[i] = 1;
			}
		}

		bleuStats.translationLength = translationLength;

		// Now go back through all n-grams in the translation and get counts of matches.
		map<string, int>::const_iterator ngramIter = translationNGrams.begin();
		map<string, int>::const_iterator refIter;
		for (; ngramIter != translationNGrams.end(); ++ngramIter) {
			// Get the count of the n-gram in the translation.
			int transCount = ngramIter->second;
			// Get the number of tokens in the n-gram.
			int ngramLength = countChar(ngramIter->first, ' ') + 1;
			if (ngramLength <= 0 || ngramLength > m_ngramOrder) {
				cerr << "Error: invalid n-gram length: " << ngramLength << endl;
			} else {
				bleuStats.ngramTotals[ngramLength-1] += transCount;

				// Search for the current n-gram in the reference.
				refIter = m_referenceLineNGrams[0].find(ngramIter->first);
				if (refIter != m_referenceLineNGrams[0].end()) {
					int refCount = refIter->second;
					if (refCount >= transCount) {
						bleuStats.ngramMatches[ngramLength-1] += transCount;
					} else {
						// There are more occurrences in the translation than in the reference,
						// so we only get credit for the count in the reference.
						bleuStats.ngramMatches[ngramLength-1] += refCount;
					}
				}
			}
		}

		double bleu = 0.0;

		vector<double> ngramPrecisions(m_ngramOrder, 0.0);

		for (int i = 0; i < m_ngramOrder; ++i) {
			if (bleuStats.ngramTotals[i] > 0) {
				ngramPrecisions[i] = ((double) bleuStats.ngramMatches[i]) / ((double) bleuStats.ngramTotals[i]);
			} else {
				ngramPrecisions[i] = 0.0;
			}
		}

		double brevityPenalty = 1.0;
		int closestRefLen = m_multiReferenceLengths[0][0];
		// Find the closest reference length if there are multiple references.
		if (m_numReferences > 1) {
			// Find the reference sentence with length closest to the translation.
			int minAbsLenDiff = 99999;
			closestRefLen = -1;
			for (unsigned int i = 0; i < m_multiReferenceLengths[0].size(); ++i) {
				int absLenDiff = abs(translationLength - m_multiReferenceLengths[0][i]);
				bool makeUpdate = false;
				if (absLenDiff < minAbsLenDiff) {
					makeUpdate = true;
				}
				if (absLenDiff == minAbsLenDiff && m_multiReferenceLengths[0][i] < closestRefLen) {
					makeUpdate = true;
				}

				if (makeUpdate) {
					minAbsLenDiff = absLenDiff;
					closestRefLen = m_multiReferenceLengths[0][i];
				}
			}
		}
		
		// If we're using the version of BLEU+1 with the fix for the brevity penalty, add 1 to the closest reference length.
		if (m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) ++closestRefLen;
		// Compute the brevity penalty using the closest reference sentence length.
		brevityPenalty = MTEval::computeBrevityPenalty(translationLength, closestRefLen);

		// Set the reference length in the BLEU sufficient stats object to the closest reference length.
		bleuStats.referenceLength = closestRefLen;

		//if (translationLength < closestRefLen) {
		//	brevityPenalty = exp(1.0 - (((double) closestRefLen) / ((double) translationLength)));
		//}

		bleu = brevityPenalty;
		double precTerm = 0.0;
		double groundedPrecTerm = 0.0;
		for (int i = 0; i < m_ngramOrder; ++i) {
			if (ngramPrecisions[i] == 0.0) {
				// If any of the n-gram precisions is 0.0, the BLEU score is 0.0.
				bleu = 0.0;
				break;
			} else {
				precTerm += log(ngramPrecisions[i]);
			}
			if (m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {
				if (bleuStats.ngramTotals[i] > 0) {
					groundedPrecTerm += log(1.0 / ((double) bleuStats.ngramTotals[i]));
				}
			}
		}
		double fullPrecTerm = exp(precTerm / 4);
		if (m_metric == BLEU_PLUS_ONE_ALL_N_PGROUND || m_metric == BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND) {
			fullPrecTerm -= exp(groundedPrecTerm / 4);
		}
		bleu *= fullPrecTerm;

		//char scoreStringFormat[2000];
		//printf("BLEU = %.2f, %.1f/%.1f/%.1f/%.1f (BP=%.3f, ratio=%.3f, hyp_len=%d, ref_len=%d)\n",
		//	100*bleu, 100*ngramPrecisions[0], 100*ngramPrecisions[1], 100*ngramPrecisions[2], 100*ngramPrecisions[3],
		//	brevityPenalty, (((double) translationLength) / ((double) closestRefLen)),
		//	translationLength, closestRefLen);
		//printScore = scoreStringFormat;
		return bleu;
	}
	return 0.0;
}

void MTEval::fillReferenceNGrams(const string& reference, map<string, int>& referenceNGrams, int* refLength) {
	*refLength = 0;
	if (m_ngramOrder > 0) {
		referenceNGrams.clear();

		vector<string> tokens;
		//splitString(reference + " ", " ", tokens);
		tokenize(reference, ' ', tokens);
		unsigned int numTokens = tokens.size();
		*refLength = numTokens;
		// Go through all tokens in the reference
		for (unsigned int i = 0; i < numTokens; ++i) {
			// Add the ith token as a unigram.
			string unigram = tokens[i];
			referenceNGrams[unigram] += 1;
			// If there is another token past the ith, add the bigram.
			if (m_ngramOrder > 1 && i < numTokens - 1) {
				string bigram = unigram + " " + tokens[i+1];
				referenceNGrams[bigram] += 1;
				// If there is another token past the (i+1)th, add the trigram.
				if (m_ngramOrder > 2 && i < numTokens - 2) {
					string trigram = bigram + " " + tokens[i+2];
					referenceNGrams[trigram] += 1;
					// If there is another token past the (i+2)th, add the fourgram.
					if (m_ngramOrder > 3 && i < numTokens - 3) {
						string fourgram = trigram + " " + tokens[i+3];
						referenceNGrams[fourgram] += 1;
					}
				}
			}
		}
	}
}

// Will insert n-gram counts from newReferenceNGrams into cumReferenceNGrams, obeying the max-count constraint
// used when computing BLEU. That is, an (n-gram, count) pair will only be inserted if the count is greater than
// its count in cumReferenceNGrams, or if it does not appear at all in cumReferenceNGrams.
void MTEval::mergeReferenceNGrams(map<string, int>& cumReferenceNGrams, const map<string, int>& newReferenceNGrams) {
	map<string, int>::const_iterator newIter;
	map<string, int>::iterator oldIter;
	for (newIter = newReferenceNGrams.begin(); newIter != newReferenceNGrams.end(); ++newIter) {
		oldIter = cumReferenceNGrams.find(newIter->first);

		if (oldIter == cumReferenceNGrams.end()) {
			// If we didn't find the n-gram, insert it.
			cumReferenceNGrams.insert(*newIter);
		} else {
			// We found the n-gram in the old count map.
			// If the new count is larger than the old one,
			if (newIter->second > oldIter->second) {
				// Insert the new count into the old entry.
				oldIter->second = newIter->second;
			}
		}
	}
}

void MTEval::fillNGramCountsFromTokens(const vector<string>& tokens, map<string, int>& ngramCounts) {
	// Go through all tokens in the translation.
	unsigned int numTokens = tokens.size();
	for (unsigned int i = 0; i < numTokens; ++i) {
		// Add the ith token as a unigram.
		string unigram = tokens[i];
		ngramCounts[unigram] += 1;
		// If there is another token past the ith, add the bigram.
		if (m_ngramOrder > 1 && i < numTokens - 1) {
			string bigram = unigram + " " + tokens[i+1];
			ngramCounts[bigram] += 1;
			// If there is another token past the (i+1)th, add the trigram.
			if (m_ngramOrder > 2 && i < numTokens - 2) {
				string trigram = bigram + " " + tokens[i+2];
				ngramCounts[trigram] += 1;
				// If there is another token past the (i+2)th, add the fourgram.
				if (m_ngramOrder > 3 && i < numTokens - 3) {
					string fourgram = trigram + " " + tokens[i+3];
					ngramCounts[fourgram] += 1;
				}
			}
		}
	}
}
