// Copyright 2010-2012 Kevin Gimpel
// Licensed under the modified BSD license. See the LICENSE file included with this distribution.

#include "pbmodel.h"

// Reads in a file containing k-best lists for all tuning sentences for a single iteration of batch training.
// kbestLists will be cleared and filled with the k-best list for each sentence.
// Returns number of lines read.
int PBModel::readKBestFile(const string& kBestListFile, vector<vector<TRANSLATION> >& kbestLists,
		int numSentences, bool readingPrevKBestLists, bool fastMode, const vector< vector<string> >& refSentences,
		const vector< vector<string> >& refSentencesRotated, 
		int currIterationNumber, const string& extraString) {
	// Format of k-best lists:
	// Hiero:
	//0 ||| karachi staff reporter ) on monday , four people including a woman were killed in ...
	//different incidents in karachi and found the body of a man from the sea ....
	//||| lm: -42.1559 tm: -23.6095 -45.966 -21.1614 -36.9725 13.9985 3.99959 w: -13.8974 ||| -10.4871
	// Moses:
	//0 ||| karachi staff reporter ) ( in karachi on monday , four people including a woman were...
	//killed in different incidents , while the body of a man from the sea ....
	//||| d: -58 -2.75094 -8.37068 -3.42464 -3.39002 -9.39276 -5.67696 lm: -106.649 w: -31...
	//tm: -28.8073 -43.7762 -20.6663 -35.2134 15.9983 ||| -7.90096

    ifstream inputFilestream;
    inputFilestream.open(kBestListFile.c_str());
    if (!inputFilestream) {
        cerr << "Error: Cannot open k-best list file " << kBestListFile << ". Exiting..." << endl;
		exit(1);
    }

	map<unsigned int, vector<TRANSLATION> > allTranslations;
	map<unsigned int, vector<TRANSLATION> >::iterator sentenceIter;

	struct timeval initialTime, time1;
	gettimeofday(&initialTime, 0);

	vector<string> linetoks;
	string str1;
	int lineCounter = 0;
	cerr << "Reading lines from ";
	if (readingPrevKBestLists) cerr << "previous ";
	cerr << "k-best list file." << endl;
	int previousSentenceIndex = 0;

	// We have one entry for each reference file, and for each, a vector holding a single line.
	vector< vector<string> > currLineMultiReferenceLines;
	//vector<string>& refSentencesi = refSentencesRotated[0];
	// refSentencesi.size() is the number of references.
	for (unsigned int ii = 0; ii < refSentencesRotated[0].size(); ++ii) {
		vector<string> currRefSingleLine;
		currRefSingleLine.push_back(refSentencesRotated[0][ii]);
		currLineMultiReferenceLines.push_back(currRefSingleLine);
	}
	//SmoothedSentenceBLEU smoothedSentenceBLEU = BLEU_PLUS_ONE;
	// Compute sentence-level BLEU, then throw away the translation string (don't store it) to save memory.
	MTEval mteval(currLineMultiReferenceLines, m_costMetric);

	vector<string> modelBestLines;
	double sumOfModelBestSentenceBLEUScores = 0.0;

    while (!inputFilestream.eof()) {
		// Get a line from the k-best list file.
    	getline(inputFilestream, str1);
		if (!inputFilestream.eof()) {
			++lineCounter;
			if (lineCounter % 400 == 0) {
				cerr << ".";
				if (lineCounter % 40000 == 0) {
					cerr << lineCounter;
					gettimeofday(&time1, 0);
					double partialTime = computeTimeDifference(initialTime, time1);
					cerr << "\tcumulative time:\t" << partialTime << endl;
				}
			}
			// Tokenize the line around " ||| ".
			linetoks.clear();
			linetoks.reserve(4);
			splitString(str1, "|||", linetoks);

			if (linetoks.size() == 4) {
				// The first token is the sentence index.
				string trimmedToken0 = trimString(linetoks[0]);
				int currentSentenceIndex = atoi(trimmedToken0.c_str());

				TRANSLATION t;
				t.featureValues.set_empty_key(999999);
				// Only store the actual translation if we are not reading in previous k-best lists in fast mode.
				//if (!(readingPrevKBestLists && fastMode)) {
				if (!fastMode) {
					// When using lower-memory translations, we need to still be able to set this.. how?
					t.text = trimString(linetoks[1]);
				} else {
					// If we are reading a translation for a different source sentence from last time, generate
					// a new set of references for computing BLEU.
					if (previousSentenceIndex != currentSentenceIndex) {
						//vector<string>& refSentencesi = refSentencesRotated[currentSentenceIndex];
						currLineMultiReferenceLines.clear();
						// refSentencesi.size() is the number of references.
						for (unsigned int ii = 0; ii < refSentencesRotated[currentSentenceIndex].size(); ++ii) {
							vector<string> currRefSingleLine;
							currRefSingleLine.push_back(refSentencesRotated[currentSentenceIndex][ii]);
							currLineMultiReferenceLines.push_back(currRefSingleLine);
						}
						mteval.updateMultiReferenceLines(currLineMultiReferenceLines);
					}
					// Compute sentence-level BLEU scores for the translation.
					// Create a dummy BLEU sufficient statistics variable that will get filled in when sentence BLEU is computed.
					BLEUSuffStats bleuSuffStats;
					t.bleuScore = mteval.computeSentenceBLEU(trimString(linetoks[1]), bleuSuffStats);
					// If we're not reading in previous k-best lists and we're in fast mode, save the first translation
					// we see for each sentence index, for purposes of computing the corpus-level BLEU score of the model-best outputs.
					if (!readingPrevKBestLists) {
						// If we are seeing a translation for a new sentence on this line, or if we're on the first line in the file, 
						// save the translation.
						if (previousSentenceIndex != currentSentenceIndex || lineCounter == 1) {
							modelBestLines.push_back(trimString(linetoks[1]));
							// Also add the sentence-level BLEU score to the sum for printing the average down below.
							sumOfModelBestSentenceBLEUScores += t.bleuScore;
						}
					}
				}
				string trimmedToken3 = trimString(linetoks[3]);
				t.score = atof(trimmedToken3.c_str());

				// Parse the feature value string.
				vector<string> featureValueTokens;
				string trimmedToken2 = trimString(linetoks[2]);
				tokenize(trimmedToken2, ' ', featureValueTokens);

				if (featureValueTokens.size() > 1) {

					// Go through all the feature value tokens. Feature types should end in a ":" and
					// be followed by a sequence of doubles containing feature values for that feature
					// type.
					for (unsigned int i = 0; i < featureValueTokens.size() - 1; ++i) {	// End at one before the last token since we are only looking for feature types.
						char lastChar = featureValueTokens[i].at(featureValueTokens[i].size() - 1);
						if (lastChar == ':') {
							string featureTypeFromKBestList = featureValueTokens[i].substr(0, featureValueTokens[i].size() - 1);
							// Look up the feature type.
							//map<string, pair<int, int> >::iterator iter1;
							dense_hash_map<string, pair<int, int>, MurmurHash2, eqstring>::iterator iter1;
							iter1 = m_featureIndex.find(featureTypeFromKBestList);
							if (iter1 == m_featureIndex.end()) {
								// If we didn't find the feature name in the moses ini list, check to see if we have a mapping for it.
								map<string, string>::iterator iter2;
								iter2 = m_kBestFeatureTypeToMosesIniFeatureType.find(featureTypeFromKBestList);
								if (iter2 == m_kBestFeatureTypeToMosesIniFeatureType.end()) {
									cerr << "Error: feature " << featureTypeFromKBestList << " in k-best list feature vector was "
										 << "not found in Moses.ini file, nor was a mapping provided for it in a mapping file." << endl;
								} else {
									// There is a mapping, so look up the mapped feature name in the moses ini list.
									iter1 = m_featureIndex.find(iter2->second);
									if (iter1 == m_featureIndex.end()) {
										cerr << "Error: feature " << featureTypeFromKBestList << " in k-best list feature vector was "
											 << "not found in Moses.ini file, nor was its mapping (" << iter2->second << ") found." << endl;
									} else {
										// Here, the mapping was found.
										processKBestListFeatureType(featureValueTokens, i+1, iter1->second.first, iter1->second.second, t.featureValues);
										i += iter1->second.second;
									}
								}
							} else {
								// We found the same feature name in the moses ini feature list, so process the feature value tokens.
								processKBestListFeatureType(featureValueTokens, i+1, iter1->second.first, iter1->second.second, t.featureValues);
								i += iter1->second.second;
							}
						}
					}
				} else {
					cerr << "Error: in k-best list file, feature value section contains " << featureValueTokens.size()
						 << " tokens. Expected more. Section: " << linetoks[2] << endl;
				}

				// Now that we have created a Translation from the line, figure out where to insert it.

				if (!readingPrevKBestLists) {
					// Check if we've seen a translation for this input sentence yet.
					sentenceIter = allTranslations.find(currentSentenceIndex);
					if (sentenceIter != allTranslations.end()) {
						sentenceIter->second.push_back(t);
					} else {
						// If the translation is for a new sentence, create a new vector of translations for the new sentence.
						vector<TRANSLATION> newKBestList;
						newKBestList.push_back(t);
						// Insert the new vector with the sentence index into the map
						allTranslations.insert(make_pair(currentSentenceIndex, newKBestList));
					}
				} else {
					// We are reading in previous k-best lists.
					// Append the translation to the vector for the current sentence.
					if (currentSentenceIndex < kbestLists.size()) {
						kbestLists[currentSentenceIndex].push_back(t);
					} else {
						cerr << "Error: set of previous k-best lists contains entries for an unknown sentence index: "
							 << currentSentenceIndex << ". (Number of source sentences = " << kbestLists.size()
							 << ".) Not adding entry." << endl;
					}
					if (fastMode && m_loss == PRO && currentSentenceIndex != previousSentenceIndex) {
						rescoreKBestList(kbestLists[previousSentenceIndex], false);
						// If we're running PRO in fast mode and we just concluded a sentence's translations, 
						// generate PRO training pairs for the sentence.
						sampleAndGenerateTrainingPairsForPRO(kbestLists[previousSentenceIndex], previousSentenceIndex);
						// Since we don't need the translations anymore, clear the k-best list to save memory.
						kbestLists[previousSentenceIndex].clear();
					}
				}
				previousSentenceIndex = currentSentenceIndex;
			} else {
				cerr << "Error: in k-best list file, line contains " << linetoks.size()
					 << " tokens. Expected 4 tokens. Line: " << str1 << endl;
			}
		}
	}
    inputFilestream.close();

	if (lineCounter > 0) {
		if (fastMode && m_loss == PRO && readingPrevKBestLists) {// && currentSentenceIndex != previousSentenceIndex) {
			rescoreKBestList(kbestLists[previousSentenceIndex], false);
			// If we're running PRO in fast mode, generate PRO training pairs for the final sentence.
			sampleAndGenerateTrainingPairsForPRO(kbestLists[previousSentenceIndex], previousSentenceIndex);
			kbestLists[previousSentenceIndex].clear();
		}
	}

	cerr << "Done reading lines from ";
	if (readingPrevKBestLists) cerr << "previous ";
	cerr << "k-best list file.";
	gettimeofday(&time1, 0);
	double finalTime = computeTimeDifference(initialTime, time1);
	cerr << "\ttime:\t" << finalTime << endl;

	if (!readingPrevKBestLists) {
		// Only initialize the given k-best lists vector if we are reading k-best lists from the current iteration.
		kbestLists.clear();
		kbestLists.reserve(numSentences);
		//kbestLists.clear();
		//kbestLists
		// Since map maintains sorted keys, this loop will process the sentence indices in order starting from 0.
		for (sentenceIter = allTranslations.begin(); sentenceIter != allTranslations.end(); ++sentenceIter) {
			// If there was a gap between the current sentence index and the previously-inserted k-best list,
			// insert an empty k-best list into the vector of k-best lists.
			while (sentenceIter->first > kbestLists.size()) {
				vector<TRANSLATION> emptyKBestList;
				kbestLists.push_back(emptyKBestList);
			}
			if (sentenceIter->first == kbestLists.size()) {
				kbestLists.push_back(sentenceIter->second);
			} else {
				// Ensure that we are inserting the sentence index at the place where we should.
				cerr << "Error: sentenceIter->first (" << sentenceIter->first << ") != kbestLists.size() ("
					 << kbestLists.size() << ")!" << endl;
			}
		}
		// Insert empty k-best lists for any remaining source sentences for which we haven't seen any translations.
		while (kbestLists.size() < numSentences) {
			vector<TRANSLATION> emptyKBestList;
			kbestLists.push_back(emptyKBestList);
		}
		allTranslations.clear();
		// If we're in fast mode, compute the corpus-level BLEU score using the first translation for each sentence.
		if (fastMode) {
			MTEval mteval(refSentences);
			string modelBestBLEUString;
			int prevIterationNumber = currIterationNumber - 1;
			mteval.computeBLEUString(modelBestLines, modelBestBLEUString);
			//cerr << "size = " << modelBestLines.size() << ", sum = " << sumOfModelBestSentenceBLEUScores << endl; 
			cerr << prevIterationNumber << "\t" << extraString << "\tsystem avg BLEU+1 = " << (sumOfModelBestSentenceBLEUScores / ((double) modelBestLines.size()))
				<< " corp " << modelBestBLEUString << endl;
		}
	}
	return lineCounter;
}


void PBModel::processKBestListFeatureType(vector<string>& tokens, int tokenIndex, int featureIndexInV, int numFeatures,
	FEAT_HASH& featureValues) {

	// Go through numFeatures tokens in the tokens vector, starting at tokenIndex.
	unsigned int last = tokenIndex + numFeatures;
	if (last > tokens.size()) {
		cerr << "Error: incorrect number of feature values in k-best list feature vector. Expected " << (tokenIndex + numFeatures)
			 << " features, but only found " << tokens.size() << "." << endl;
		last = tokens.size();
	}
	for (unsigned int i = tokenIndex; i < last; ++i) {
		char lastChar = tokens[i].at(tokens[i].size() - 1);
		if (lastChar == ':') {
			cerr << "Error: encountered feature type in k-best list feature vector while trying to extract " << numFeatures << " feature values." << endl;
		} else {
			featureValues[featureIndexInV + (i - tokenIndex)] = atof(tokens[i].c_str());
		}
	}
}

// Searches k-best list for optimal translation under a variety of scoring functions, including
// model score, model + cost, model - cost, -model - cost, etc.
// Returns index in k-best list of optimal translation.
int PBModel::findOptimalTranslation(vector<TRANSLATION>& kbestList, bool findMax, bool costAugmented,
		double costMultiplier, bool useModelScoreForMax) {
	int optimalIndex = 0;
	if (kbestList.size() > 0) {
		double optimalValue = 0.0;
		if (useModelScoreForMax) optimalValue = kbestList[0].score;
		if (costAugmented) optimalValue += (costMultiplier - (costMultiplier * kbestList[0].bleuScore));

		for (unsigned int i = 0; i < kbestList.size(); ++i) {
			double currValue = 0.0;
			if (useModelScoreForMax) currValue = kbestList[i].score;
			if (costAugmented) currValue += (costMultiplier - (costMultiplier * kbestList[i].bleuScore));
			if (findMax) {
				if (currValue > optimalValue) {
					optimalValue = currValue;
					optimalIndex = i;
				}
			} else {
				if (currValue < optimalValue) {
					optimalValue = currValue;
					optimalIndex = i;
				}
			}
		}
	}
	return optimalIndex;
}

// Recomputes the score of each translation in the given k-best list using current feature weights.
// The k-best list is NOT re-sorted.
void PBModel::rescoreKBestList(vector<TRANSLATION>& kbestList, bool logDifferences) {
	for (unsigned int i = 0; i < kbestList.size(); ++i) {
		// Recompute the model score from the feature vector and the parameters.
		double newScore = 0.0;
		FEAT_HASH::iterator iter;
		for (iter = kbestList[i].featureValues.begin(); iter != kbestList[i].featureValues.end(); ++iter) {
			newScore += (getValue(iter->first) * (iter->second));
		}
		// If the newly-computed score is more than 0.1 off from the previous score, and we're
		// logging such differences, then print a warning.
		if (logDifferences && (kbestList[i].score < newScore - 0.1 || kbestList[i].score > newScore + 0.1)) {
			cerr << "Warning: kbestList[" << i << "].score = " << kbestList[i].score << ", newScore = " << newScore << endl;
		}
		kbestList[i].score = newScore;
	}
}

void PBModel::setLossFunction(const string& lossFunction) {
	m_loss = RAMP3;
	if (lossFunction == "ramp1") {
		m_loss = RAMP1;
	} else if (lossFunction == "ramp2") {
		m_loss = RAMP2;
	} else if (lossFunction == "ramp3") {
		m_loss = RAMP3;
	} else if (lossFunction == "softramp1") {
		m_loss = SOFTRAMP1;
	} else if (lossFunction == "softramp2") {
		m_loss = SOFTRAMP2;
	} else if (lossFunction == "softramp3") {
		m_loss = SOFTRAMP3;
	} else if (lossFunction == "hardsoftramp") {
		m_loss = HARDSOFTRAMP;
	} else if (lossFunction == "hardsoftramp_och") {
		m_loss = HARDSOFTRAMP_OCH;
	} else if (lossFunction == "hardsoftramp_smm") {
		m_loss = HARDSOFTRAMP_SMM;
	} else if (lossFunction == "bayesrisk") {
		m_loss = BAYESRISK;
	} else if (lossFunction == "allpairs_hinge") {
		m_loss = ALLPAIRS_HINGE;
	} else if (lossFunction == "allpairs_softhinge") {
		m_loss = ALLPAIRS_SOFTHINGE;
	} else if (lossFunction == "exp") {
		m_loss = EXPLOSS;
	} else if (lossFunction == "allpairs_exp") {
		m_loss = ALLPAIRS_EXP;
	} else if (lossFunction == "allpairs_expcost") {
		m_loss = ALLPAIRS_EXPCOST;
	} else if (lossFunction == "linear_regression") {
		m_loss = LINEAR_REGRESSION;
	} else if (lossFunction == "pro") {
		m_loss = PRO;
	} else {
		cerr << "Error while setting loss function. Unexpected loss function: " << lossFunction << ". Using RAMP3." << endl;
	}
}

void PBModel::computeBLEUScoresForKBestLists(const vector< vector<string> >& refSentences,
		const vector< vector<string> >& refSentencesRotated, int currIterationNumber, const string& extraString) {
	// Compute BLEU scores for all entries in all k-best lists.
	vector<string> oracleLines, modelBestLines;
	double sumOfOracleSentenceBLEUScores = 0.0;
	double sumOfModelBestSentenceBLEUScores = 0.0;
	// We have one entry for each reference file, and for each, a vector holding a single line.
	vector< vector<string> > currLineMultiReferenceLines;
	bool findMaxTrue = true;
	bool costAugmentedFalse = false;
	bool useModelScoreForMaxTrue = true;

	for (unsigned int i = 0; i < m_kbestLists.size(); ++i) {
		//vector<string>& refSentencesi = refSentencesRotated[i];
		currLineMultiReferenceLines.clear();
		// refSentences.size() is the number of references.
		for (unsigned int ii = 0; ii < refSentencesRotated[i].size(); ++ii) {
			vector<string> currRefSingleLine;
			currRefSingleLine.push_back(refSentencesRotated[i][ii]);
			currLineMultiReferenceLines.push_back(currRefSingleLine);
		}
		// Compute sentence-level BLEU scores for all translations on the k-best list.
		int bestBLEUIndex = computeSentenceLevelBLEUScores(m_kbestLists[i], currLineMultiReferenceLines);

		// Add the line corresponding to the best BLEU score to the vector of oracle lines.
		// Comment out .text lines for faster mode.
		oracleLines.push_back(m_kbestLists[i][bestBLEUIndex].text);
		sumOfOracleSentenceBLEUScores += m_kbestLists[i][bestBLEUIndex].bleuScore;

		// Find the highest-scoring translation under the model.
		int modelBestIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedFalse, 0.0, useModelScoreForMaxTrue);

		// Add the line corresponding to the best model score to the vector of model-best lines.
		modelBestLines.push_back(m_kbestLists[i][modelBestIndex].text);
		sumOfModelBestSentenceBLEUScores += m_kbestLists[i][modelBestIndex].bleuScore;
	}

	// Compute corpus-level BLEU score of both model-best and oracle translations for each sentence.
	MTEval mteval(refSentences);
	string oracleBLEUString, modelBestBLEUString;
	mteval.computeBLEUString(oracleLines, oracleBLEUString);
	int prevIterationNumber = currIterationNumber - 1;
	cerr << prevIterationNumber << "\t" << extraString << "\toracle avg BLEU+1 = " << (sumOfOracleSentenceBLEUScores / ((double) oracleLines.size()))
		<< " corp " << oracleBLEUString << endl;
	mteval.computeBLEUString(modelBestLines, modelBestBLEUString);
	cerr << prevIterationNumber << "\t" << extraString << "\tsystem avg BLEU+1 = " << (sumOfModelBestSentenceBLEUScores / ((double) modelBestLines.size()))
		<< " corp " << modelBestBLEUString << endl;
}

double PBModel::computeBLEUScoresForKBestListsAfterParameterUpdate(const vector< vector<string> >& refSentences,
		int currIterationNumber, int currCCCPIterationNumber, int subIterationNumber) {
	// Compute BLEU scores for the set of model-best entries across the k-best lists.
	vector<string> modelBestLines;
	double sumOfModelBestSentenceBLEUScores = 0.0;
	bool findMaxTrue = true;
	bool costAugmentedFalse = false;
	bool useModelScoreForMaxTrue = true;

	for (unsigned int i = 0; i < m_kbestLists.size(); ++i) {
		// Since a parameter update was made, rescore the current k-best list.
		rescoreKBestList(m_kbestLists[i], false);

		// Find the highest-scoring translation.
		int modelBestIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedFalse, 0.0, useModelScoreForMaxTrue);

		// Add the line corresponding to the best model score to the vector of model-best lines.
		modelBestLines.push_back(m_kbestLists[i][modelBestIndex].text);
		sumOfModelBestSentenceBLEUScores += m_kbestLists[i][modelBestIndex].bleuScore;
	}

	// Compute corpus-level BLEU score of both model-best and oracle translations for each sentence.
	MTEval mteval(refSentences);
	string modelBestBLEUString;
	double bleu = mteval.computeBLEUString(modelBestLines, modelBestBLEUString);
	cerr << "iter_" << currIterationNumber << "\tcccp_" << currCCCPIterationNumber << "\tsub_" << subIterationNumber
		<< "\tavg BLEU+1:\t" << (sumOfModelBestSentenceBLEUScores / ((double) modelBestLines.size())) << "\tcorp BLEU:\t" << bleu << endl;
		//<< " corp " << modelBestBLEUString << endl;
	return bleu;
}

// What follows is the new training function for batch training algorithms that take in k-best lists
// from a single iteration and the parameters used to generate those k-best lists and write new
// Moses config files with the result of a single iteration of parameter updates. Alternatively,
// lattices can be provided for decoding with non-local features or decoding by phrase dependency parsing.
// In all cases, k-best lists are obtained and used for training.
void PBModel::train(const string& srcSentsFile, const string& kBestFile, const string& prevKBestFile, const string& refSentsFile,
		const string& outMosesIniFile, unsigned int currIterationNumber, unsigned int numCCCPIters, unsigned int numSubIters,
		const string& lossFunction, const string& costFunctionMetric, const string& optimizationAlgorithm,
		double costCoefficient, double stepsize, double l2Reg, bool fastMode) {

	// Set the loss function
	setLossFunction(lossFunction);
	// Set the optimization algorithm.
	OptimizationAlgorithm optAlg = SSD;
	if (optimizationAlgorithm == "sgd" || optimizationAlgorithm == "ssd") {
		optAlg = SSD;
	} else if (optimizationAlgorithm == "mira") {
		optAlg = MIRA;
	} else if (optimizationAlgorithm == "pegasos") {
		optAlg = PEGASOS;
	} else {
		cerr << "Error: unexpected optimization algorithm \"" << optimizationAlgorithm
			 << "\". Please specify \"ssd\", \"sgd\", \"mira\", or \"pegasos\". Using SGD/SSD." << endl;
		optAlg = SSD;
	}
	//enum Metric {STANDARD_BLEU, LIN_OCH_BLEU_PLUS_ONE, BLEU_PLUS_ONE_ALL_N, BLEU_PLUS_ONE_ALL_N_PGROUND, BLEU_PLUS_ONE_ALL_N_AND_BP, BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND};
	if (costFunctionMetric == "STANDARD_BLEU") {
		m_costMetric = STANDARD_BLEU;
	} else if (costFunctionMetric == "LIN_OCH_BLEU_PLUS_ONE") {
		m_costMetric = LIN_OCH_BLEU_PLUS_ONE;
	} else if (costFunctionMetric == "BLEU_PLUS_ONE_ALL_N") {
		m_costMetric = BLEU_PLUS_ONE_ALL_N;
	} else if (costFunctionMetric == "BLEU_PLUS_ONE_ALL_N_PGROUND") {
		m_costMetric = BLEU_PLUS_ONE_ALL_N_PGROUND;
	} else if (costFunctionMetric == "BLEU_PLUS_ONE_ALL_N_AND_BP") {
		m_costMetric = BLEU_PLUS_ONE_ALL_N_AND_BP;
	} else if (costFunctionMetric == "BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND") {
		m_costMetric = BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND;
	} else {
		cerr << "Error: unexpected metric for cost function \"" << costFunctionMetric
			 << "\". Please specify \"STANDARD_BLEU\", \"LIN_OCH_BLEU_PLUS_ONE\", "
			 << "\"BLEU_PLUS_ONE_ALL_N\", \"BLEU_PLUS_ONE_ALL_N_PGROUND\","
			 << "\"BLEU_PLUS_ONE_ALL_N_AND_BP\", or \"BLEU_PLUS_ONE_ALL_N_AND_BP_PGROUND\". Using BLEU_PLUS_ONE_ALL_N_AND_BP." << endl;
		m_costMetric = BLEU_PLUS_ONE_ALL_N_AND_BP;
	}

	// Set some member variables for use with the Eval() function used by LBFGS.
	m_numSubIters = numSubIters;

	bool logDomainFalse = false;
	bool findMaxTrue = true;
	bool findMaxFalse = false;
	bool costAugmentedTrue = true;
	bool costAugmentedFalse = false;
	bool useModelScoreForMaxTrue = true;

	bool readingPrevKBestListsTrue = true;
	bool readingPrevKBestListsFalse = false;

	cerr << "Starting training." << endl;
	struct timeval initialTime, time1, time2;
	gettimeofday(&initialTime, 0);

	vector<string> srcSentences;
	// Read in the source lines from the source file.
	readLinesFromFile(srcSentsFile, srcSentences);
	if (srcSentences.size() == 0) {
		cerr << "Empty or missing source sentence file. Exiting..." << endl;
	}

	// Read in the reference translations from (possibly) multiple reference files.
	vector<vector<string> > refSentences;	// Outer vector is over reference files, inner vector is over lines.
	loadReferences(refSentsFile, refSentences);
	int numReferences = refSentences.size();
	if (numReferences == 0) {
		cerr << "Error: zero references given! Exiting..." << endl;
		exit(1);
	}
	unsigned int numReferenceLines = refSentences[0].size();
	for (unsigned int i = 1; i < refSentences.size(); ++i) {
		if (refSentences[i].size() != numReferenceLines) {
			cerr << "Error: number of reference sentences in reference file " << i << " = " << refSentences[i].size()
				 << " but reference file 0 has " << numReferenceLines << " lines. Exiting..." << endl;
			exit(1);
		}
	}
	if (srcSentences.size() != numReferenceLines) {
		cerr << "Error: number of source sentences: " << srcSentences.size() << ", number of reference sentences: "
			 << numReferenceLines << ". Exiting..." << endl;
		exit(1);
	}
	// Rotate them (take the transpose of the matrix) for convenience below.
	vector< vector<string> > refSentencesRotated;	// Outer vector is over lines, inner vector is over reference files.
	rotateTwoDimensionalVector(refSentences, refSentencesRotated);

	// We have one entry for each reference file, and for each, a vector holding a single line.
	vector< vector<string> > currLineMultiReferenceLines;

	// Read in the k-best lists from the k-best list file. They should be numbered from 0 up to srcSentences.size()-1. Missing indices
	// are ok. They can occur when no translations were found for a given input sentence.
	int numLinesRead = readKBestFile(kBestFile, m_kbestLists, srcSentences.size(), readingPrevKBestListsFalse, fastMode, refSentences, 
		refSentencesRotated, currIterationNumber, "onlycurr_kbest_fast");
	// Rescore the k-best lists. This is done because the provided scores include the unknown 
	// word penalty, which is not apparent from the feature vectors. By re-computing the scores
	// of the translations, we only take into account the features shown in the k-best list file.
	// This also ensures that the scores are accurately reflected by the features and current weights.
	gettimeofday(&time1, 0);
	for (unsigned int i = 0; i < m_kbestLists.size(); ++i) {
		rescoreKBestList(m_kbestLists[i], false);
	}
	gettimeofday(&time2, 0);
	double rescoringTime = computeTimeDifference(time1, time2);
	cerr << "Rescoring time:\t" << rescoringTime << " seconds" << endl;

	unsigned int numTrainingExamples = m_kbestLists.size();
	cerr << "Num training examples: " << numTrainingExamples << endl;
	if (numReferenceLines != numTrainingExamples) {
		cerr << "Warning: number of reference translations: " << numReferenceLines << ", number of k-best lists: " << numTrainingExamples << endl;
	}
	if (srcSentences.size() != numTrainingExamples) {
		cerr << "Warning: number of source sentences: " << srcSentences.size() << ", number of k-best lists: " << numTrainingExamples << endl;
	}

	// Compute BLEU score of model-best translations in k-best lists from most recent iteration.
	if (!fastMode) {
		gettimeofday(&time1, 0);
		computeBLEUScoresForKBestLists(refSentences, refSentencesRotated, currIterationNumber, "onlycurr_kbest");
		gettimeofday(&time2, 0);
		double computeBLEUTime = computeTimeDifference(time1, time2);
		cerr << "Compute BLEU time:\t" << computeBLEUTime << " seconds" << endl;
	}

	// Read in the previous k-best lists. They should be numbered from 0 up to srcSentences.size()-1.
	numLinesRead = readKBestFile(prevKBestFile, m_kbestLists, srcSentences.size(), readingPrevKBestListsTrue, fastMode, refSentences, 
		refSentencesRotated, currIterationNumber, "curr+prev_kbest_fast");
	// Only do the following if we read some lines from the previous k-best file.
	if (numLinesRead > 0) {
		// Rescore the k-best lists because the scores were computed using old weights.
		for (unsigned int i = 0; i < m_kbestLists.size(); ++i) {
			rescoreKBestList(m_kbestLists[i], false);
		}
		// Compute BLEU score of ALL k-best lists, including current and previous.
		// This will let us see a bound on the gap in performance due to search error.
		if (!fastMode) {
			gettimeofday(&time1, 0);
			computeBLEUScoresForKBestLists(refSentences, refSentencesRotated, currIterationNumber, "curr+prev_kbest");
			gettimeofday(&time2, 0);
			double computeBLEUTime = computeTimeDifference(time1, time2);
			cerr << "Compute BLEU time:\t" << computeBLEUTime << " seconds" << endl;
		}
	} else {
		cerr << "Previous k-best list file was empty." << endl;
	}

	// Now do training for this iteration given the set of scored k-best lists.

	// Supported loss functions:
	// ramp1, ramp2, ramp3,
	// softramp1, softramp2, softramp3,
	// hardsoftramp, hardsoftramp_och, hardsoftramp_smm,
	// bayesrisk,
	// allpairs_hinge, allpairs_softhinge,
	// exp, allpairs_exp

	unsigned int numExamples = m_kbestLists.size();

	if (m_loss == PRO) {
		bool useSigmoid = false;
		bool proSubsampling = true;

		/*if (proSubsamplingMethod == "alpha_thresh") {
			;
		} else if (proSubsamplingMethod == "sigmoid") {
			useSigmoid = true;
		} else if (proSubsamplingMethod == "none") {
			proSubsampling = false;
		} else {
			cerr << "Warning: using PRO with unexpected subsampling method: " << proSubsamplingMethod << endl;
		}*/

		// Create a binary classifier for the current batch of k-best lists.

		double lossValue = 0.0;
		double squaredNorm = 0.0;
		// If we're using fast mode, we already added the training pairs to the binary classifier when we read in the k-best lists.
		if (currIterationNumber == 2 || !fastMode) {
			for (unsigned int i = 0; i < m_kbestLists.size(); ++i) {
				//rescoreKBestList(m_kbestLists[i]);

				vector<TRANSLATION> activeKBestList;
				int kbestListStartingIndex = activeKBestList.size();
				// Add the k-best derivations to activeKBestList.
				for (unsigned int ii = 0; ii < m_kbestLists[i].size(); ++ii) {
					activeKBestList.push_back(m_kbestLists[i][ii]);
				}

				// Sample training pairs and sort according to BLEU score.
				vector<PROTranslationPair> translationPairs;
				int proGamma = 5000;
				int proXi = 50;
				double proAlphaThreshold = 0.05;
				proSampleTrainingPairs(activeKBestList, proGamma, proAlphaThreshold, useSigmoid, proSubsampling, translationPairs);
				cerr << "Sentence " << i << ": # pairs added = " << translationPairs.size() << endl;

				// Add the top xi pairs as training examples.
				bool proBothExamples = true;
				proAddTrainingPairs(activeKBestList, translationPairs, proXi, proBothExamples);
			}
		}

		// We have generated pairs for the current mini-batch, so train the classifier.
		vector<double> oldWeights = V;
		//int binaryClassifierTrainingMaxIters = 300;
		double l1Reg = 0.0;
		m_binaryClassifier.train(numSubIters, l1Reg, l2Reg);
		cerr << "Weights:" << endl;
		//double psi = 0.1;
		for (map<int, double>::iterator iter = m_binaryClassifier.m_sparseWeightVector.begin();
			iter != m_binaryClassifier.m_sparseWeightVector.end(); ++iter) {
			if (iter->second != 0.0) {
				cerr << iter->first << "\t" << getParameterName(iter->first) << "\t" << iter->second << endl;
			}
			setValue(iter->first, iter->second);
		}
		double proPsi = 0.1;
		for (unsigned int ii = 0; ii < V.size(); ++ii) {
			double oldWeight = oldWeights[ii];
			double newWeight = getValue(ii);
			setValue(ii, (oldWeight * (1.0 - proPsi)) + (newWeight * proPsi));
		}
	}

	// For smooth loss functions, whether convex or non-convex, we use SGD.
	// This category includes allpairs_softhinge (which is convex), softramp1, softramp2, softramp3, bayesrisk, and all-pairs exp.
	// We also handle the hardsoftramp losses here because once we handle the concave part of the loss, the convex part is smooth.
	// In computeLossGradientForSentence, we handle the hardsoftramp losses different from the others in this section.
	if (m_loss == ALLPAIRS_SOFTHINGE || m_loss == SOFTRAMP1 || m_loss == SOFTRAMP2 || m_loss == SOFTRAMP3 || m_loss == BAYESRISK ||
		m_loss == HARDSOFTRAMP || m_loss == HARDSOFTRAMP_OCH || m_loss == HARDSOFTRAMP_SMM || m_loss == ALLPAIRS_EXP ||
		m_loss == ALLPAIRS_EXPCOST || m_loss == LINEAR_REGRESSION) {

		// Run CCCP for numCCCPIters iterations.
		for (unsigned int cccpIter = 1; cccpIter <= numCCCPIters; ++cccpIter) {
			// If we're using one of the non-differentiable, non-convex losses, the first thing we need to do on this
			// CCCP iteration is impute values in the concave term.
			if (m_loss == HARDSOFTRAMP || m_loss == HARDSOFTRAMP_OCH || m_loss == HARDSOFTRAMP_SMM) {
				imputeConcaveTerm(costCoefficient);
			}

			// Run convex optimizer for numSubIters iterations; this might be SSD or PEGASOS.
			for (unsigned int subIter = 1; subIter <= numSubIters; ++subIter) {
				double lossValue = 0.0;
				double squaredNorm = 0.0;
				for (unsigned int i = 0; i < numExamples; ++i) {
					double currSquaredNorm = 0.0;
					rescoreKBestList(m_kbestLists[i], false);

					// Reset the gradient vector since we make an update after processing each example.
					resetGradientVector(logDomainFalse);
					// Compute and update the gradient vector for this example, also computing the loss.
					double currLossValue = computeLossGradientForSentence(i, costCoefficient);
					lossValue += currLossValue;

					double stepsizeToUse = stepsize;
					if (optAlg == PEGASOS) {
						stepsizeToUse = 1.0 / (l2Reg * ((subIter * numExamples) + (i + 1)));
					}
					// Add the gradient for the L2 regularization term.
					if (l2Reg > 0.0) {
						for (PARAM_ID j = 1; j < V.size(); ++j) {
							if (m_priorMeansGiven) {
								double priorMean = 0.0;
								if (m_priorMeans.size() > j) priorMean = m_priorMeans[j];
								// The subgradient for the regularization term with a prior mean \mu is:
								// - C(V[i] - \mu)
								// We have to scale the gradient by mini-batch size (which is always 1) / total number of training examples.
								if (optAlg == SSD) {
									G[j] = G[j] - (stepsizeToUse * l2Reg * ((V[j] - priorMean) / numTrainingExamples));
								} else if (optAlg == PEGASOS) {
									G[j] = G[j] - (stepsizeToUse * l2Reg * (V[j] - priorMean));
								}
							} else {
								// The subgradient for the regularization term with mean 0.0 is:
								// - C*V[i]
								// We have to scale the gradient by mini-batch size (which is always 1) / total number of training examples.
								if (optAlg == SSD) {
									G[j] = G[j] - (stepsizeToUse * l2Reg * (V[j] / numTrainingExamples));
								} else if (optAlg == PEGASOS) {
									G[j] = G[j] - (stepsizeToUse * l2Reg * V[j]);
								}
							}
							//	cerr << "after reg\t" << i << "\t" << j << "\tG for " << getParameterName(j) << " = " << G[j] << endl;
						}
					}

					// Go back through the gradient vectors and perform the gradient update.
					for (PARAM_ID j = 0; j < G.size(); ++j) {
						if (G[j] != G[j]) {
							cerr << "Error: NaN gradient, sentence " << i << "\tparam " << j << "\tV for " << getParameterName(j) << " = " << V[j] << endl;
						} else {
							V[j] -= (stepsizeToUse * G[j]);
							squaredNorm += (G[j] * G[j]);
							currSquaredNorm += (G[j] * G[j]);
						}
					}
				}

				// Compute the BLEU score of the reranked tuning set.
				if (!fastMode) computeBLEUScoresForKBestListsAfterParameterUpdate(refSentences, currIterationNumber-1, cccpIter, subIter);
				// Print the value of the loss, the squared norm of the loss subgradient, and the number of examples with zero loss.
				cerr << "iter_" << (currIterationNumber-1);
				if (m_loss == HARDSOFTRAMP || m_loss == HARDSOFTRAMP_OCH || m_loss == HARDSOFTRAMP_SMM) {
					cerr << "\tcccp_" << cccpIter;
				}
				cerr << "\tsub_" << subIter << "\tLoss:\t" << lossValue << "\tSquaredNorm:\t" << squaredNorm << endl;
				//cerr << currIterationNumber << "\t" << subIter << "\tLoss:\t" << lossValue << "\tSquaredGradNorm\t" << squaredNorm << endl;

			} // end loop over sub-iterations

			if (m_loss == HARDSOFTRAMP || m_loss == HARDSOFTRAMP_OCH || m_loss == HARDSOFTRAMP_SMM) {
				cerr << currIterationNumber << "\tFinished CCCP iteration\t" << cccpIter << endl;
			}
		} // end loop over CCCP iterations
	}

	// For non-smooth, convex loss functions, we use stochastic subgradient descent.
	// The only convex loss in this category is allpairs_hinge, but we can use this for the ramp losses after imputation.
	// After updating the parameters on sentence i, we simply rerank the k-best list
	// for sentence i+1, compute the subgradient, and make the update. That is, we do
	// not compute a new k-best list from Moses. We probably should, but we don't for reasons of computational efficiency.
	if (m_loss == ALLPAIRS_HINGE || m_loss == RAMP1 || m_loss == RAMP2 || m_loss == RAMP3) {

		// Run CCCP for numCCCPIters iterations.
		for (unsigned int cccpIter = 1; cccpIter <= numCCCPIters; ++cccpIter) {
			// If we're using one of the non-convex losses, the first thing we need to do on this CCCP iteration is impute values in the concave term.
			if (m_loss == RAMP1 || m_loss == RAMP2 || m_loss == RAMP3) {
				imputeConcaveTerm(costCoefficient);
			}

			// Run convex optimizer for numSubIters iterations; this might be SSD, MIRA, or PEGASOS.
			for (unsigned int subIter = 1; subIter <= numSubIters; ++subIter) {
				int numSkippedUpdates = 0;
				double lossValue = 0.0;
				double squaredNorm = 0.0;

				for (unsigned int i = 0; i < numExamples; ++i) {
					// TODO: only rescore if we definitely made an update on the previous round?
					//if (i > 0)
					rescoreKBestList(m_kbestLists[i], false);

					double stepsizeToUse = stepsize;
					if (optAlg == PEGASOS) {
						stepsizeToUse = 1.0 / (l2Reg * (((subIter-1) * numExamples) + (i + 1)));
					}

					int towardsIndex;
					int awayIndex;
					double currLoss = 0.0;
					if (m_loss == ALLPAIRS_HINGE) {
						// The translation to update towards will be the one that minimizes model+cost.
						towardsIndex = findOptimalTranslation(m_kbestLists[i], findMaxFalse, costAugmentedTrue, costCoefficient, useModelScoreForMaxTrue);
						// The translation to update away from will be the one that maximizes model+cost.
						awayIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedTrue, costCoefficient, useModelScoreForMaxTrue);
						if (!fastMode || optAlg == MIRA) {
							currLoss += (-m_kbestLists[i][towardsIndex].score - computeCost(costCoefficient, m_kbestLists[i][towardsIndex].bleuScore));
							currLoss += (m_kbestLists[i][awayIndex].score + computeCost(costCoefficient, m_kbestLists[i][awayIndex].bleuScore));
						}
					} else if (m_loss == RAMP1) {
						// The translation to update towards will be the one that maximizes model (already found above and unchanged during this iteration).
						towardsIndex = m_imputedIndices[i];
						// The translation to update away from will be the one that maximizes model+cost.
						awayIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedTrue, costCoefficient, useModelScoreForMaxTrue);

						if (!fastMode || optAlg == MIRA) {
							currLoss -= m_kbestLists[i][towardsIndex].score;
							currLoss += (m_kbestLists[i][awayIndex].score + computeCost(costCoefficient, m_kbestLists[i][awayIndex].bleuScore));
						}
					} else if (m_loss == RAMP2) {
						// The translation to update towards will be the one that maximizes model - cost (already found above and unchanged during this iteration).
						towardsIndex = m_imputedIndices[i];
						// The translation to update away from will be the one that maximizes model.
						awayIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedFalse, costCoefficient, useModelScoreForMaxTrue);

						if (!fastMode || optAlg == MIRA) {
							currLoss -= (m_kbestLists[i][towardsIndex].score - computeCost(costCoefficient, m_kbestLists[i][towardsIndex].bleuScore));
							currLoss += m_kbestLists[i][awayIndex].score;
						}
					} else if (m_loss == RAMP3) {
						// The translation to update towards will be the one that maximizes model - cost (already found above and unchanged during this iteration).
						towardsIndex = m_imputedIndices[i];
						// The translation to update away from will be the one that maximizes model+cost.
						awayIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedTrue, costCoefficient, useModelScoreForMaxTrue);

						if (!fastMode || optAlg == MIRA) {
							currLoss -= (m_kbestLists[i][towardsIndex].score - computeCost(costCoefficient, m_kbestLists[i][towardsIndex].bleuScore));
							currLoss += (m_kbestLists[i][awayIndex].score + computeCost(costCoefficient, m_kbestLists[i][awayIndex].bleuScore));
						}
					} else {
						cerr << "Error: unexpected loss function in non-smooth loss function section: " << m_loss << endl;
						exit(1);
					}
					lossValue += currLoss;
					// Add the subgradient for the L2 regularization term if the L2 coefficient is greater than 0 and we're not using MIRA.
					if (l2Reg > 0.0 && optAlg != MIRA) {
						for (PARAM_ID j = 1; j < V.size(); ++j) {
							if (m_priorMeansGiven) {
								double priorMean = 0.0;
								if (m_priorMeans.size() > j) priorMean = m_priorMeans[j];
								// The subgradient for the regularization term with a prior mean \mu is:
								// - C(V[i] - \mu)
								// We have to scale the subgradient by mini-batch-size (which is always 1) / total number of training examples.
								if (optAlg == SSD) {
									V[j] = V[j] - (stepsizeToUse * l2Reg * ((V[j] - priorMean) / ((double)numTrainingExamples)));
								} else if (optAlg == PEGASOS) {
									V[j] = V[j] - (stepsizeToUse * l2Reg * (V[j] - priorMean));
								}
							} else {
								// The subgradient for the regularization term with mean 0.0 is:
								// - C V[i]
								// We have to scale the subgradient by mini-batch-size (which is always 1) / total number of training examples.
								if (optAlg == SSD) {
									V[j] = V[j] - (stepsizeToUse * l2Reg * (V[j] / ((double)numTrainingExamples)));
								} else if (optAlg == PEGASOS) {
									V[j] = V[j] - (stepsizeToUse * l2Reg * V[j]);
								}
							}
						}
					}

					if (towardsIndex == awayIndex) {
						++numSkippedUpdates;
					} else {
						if (optAlg == MIRA) {
							double maxStepsize = l2Reg;
							if (maxStepsize <= 0.0) {
								maxStepsize = 1e-8;
							}

							double stepsizeNumerator = currLoss - computeCost(costCoefficient, m_kbestLists[i][towardsIndex].bleuScore);
							double stepsizeDenom = computeSquaredNormOfSparseFeatureMapDifference(m_kbestLists[i][awayIndex].featureValues,
								m_kbestLists[i][towardsIndex].featureValues);
							double candidateStepsize = stepsizeNumerator / stepsizeDenom;
							stepsizeToUse = min(maxStepsize, candidateStepsize);
							cerr << "Curr loss = " << currLoss << ", stepsizeNum = " << stepsizeNumerator << ", stepsizeDenom = "
								<< stepsizeDenom << ", candidate stepsize = " << candidateStepsize << ", used stepsize = " << stepsizeToUse;
							if (stepsizeToUse == maxStepsize) cerr << " (used max)";
							cerr << endl;
							if (stepsizeToUse < 0.0) {
								cerr << "Error: stepsize to use is negative: " << stepsizeToUse << ". Using " << maxStepsize << "." << endl;
								stepsizeToUse = maxStepsize;
							}
						}

						// Add the feature vector for the translation to update towards.
						addSparseFeatureValueMapToV(m_kbestLists[i][towardsIndex].featureValues, stepsizeToUse);
						// Subtract the feature vector for the translation to update away from.
						subtractSparseFeatureValueMapFromV(m_kbestLists[i][awayIndex].featureValues, stepsizeToUse);
						//printFeatureMapDifference(m_kbestLists[i][towardsIndex].featureValues, m_kbestLists[i][awayIndex].featureValues, stepsizeToUse);
						if (!fastMode) {
							double currSquaredNorm = computeSquaredNormOfSparseFeatureMapDifference(m_kbestLists[i][towardsIndex].featureValues,
								m_kbestLists[i][awayIndex].featureValues);
							squaredNorm += currSquaredNorm;
						}
					}
				}

				// Compute the BLEU score of the reranked tuning set.
				if (!fastMode) computeBLEUScoresForKBestListsAfterParameterUpdate(refSentences, currIterationNumber-1, cccpIter, subIter);
				// Print the value of the loss, the squared norm of the loss subgradient, and the number of examples with zero loss.
				cerr << "iter_" << (currIterationNumber-1) << "\tcccp_" << cccpIter << "\tsub_" << subIter;
				if (!fastMode || optAlg == MIRA) cerr << "\tLoss:\t" << lossValue;
				if (!fastMode) cerr << "\tSquaredNorm:\t" << squaredNorm;
				cerr << "\tNumZeroLoss:\t" << numSkippedUpdates << endl;

			} // end loop over sub-iterations
			cerr << currIterationNumber << "\tFinished CCCP iteration\t" << cccpIter << endl;
		} // end loop over CCCP iterations
	}

	writeNewMosesIniFile(outMosesIniFile);
	gettimeofday(&time1, 0);
	double learningTime = computeTimeDifference(initialTime, time1);
	cerr << currIterationNumber << "\tTime:\t" << learningTime << endl;
}

void PBModel::copySparseFeatureValues(const FEAT_HASH& source, map<int, double>& dest) {

	dest.clear();
	FEAT_HASH::const_iterator iter;
	for (iter = source.begin(); iter != source.end(); ++iter) {
		dest[iter->first] = iter->second;
	}
}

void PBModel::sampleAndGenerateTrainingPairsForPRO(const vector<TRANSLATION>& activeKBestList, int currentIndex) {
	//for (unsigned int i = 0; i < m_kbestLists.size(); ++i) {
		//rescoreKBestList(m_kbestLists[i]);

		/*int kbestListStartingIndex = activeKBestList.size();
		// Add the k-best derivations to activeKBestList.
		for (unsigned int ii = 0; ii < m_kbestLists[i].size(); ++ii) {
			activeKBestList.push_back(m_kbestLists[i][ii]);
		}*/

		// Sample training pairs and sort according to BLEU score.
		vector<PROTranslationPair> translationPairs;
		int proGamma = 5000;
		int proXi = 50;
		double proAlphaThreshold = 0.05;
		bool useSigmoid = false;
		bool proSubsampling = true;
		proSampleTrainingPairs(activeKBestList, proGamma, proAlphaThreshold, useSigmoid, proSubsampling, translationPairs);
		cerr << "Sentence " << currentIndex << ": # pairs added = " << translationPairs.size() << endl;

		// Add the top xi pairs as training examples.
		bool proBothExamples = true;
		proAddTrainingPairs(activeKBestList, translationPairs, proXi, proBothExamples);
	//}
}

void PBModel::proSampleTrainingPairs(const vector<TRANSLATION>& activeKBestList, int gamma,
		double proAlphaThreshold, bool useSigmoid, bool proNoSubsampling, vector<PROTranslationPair>& translationPairs) {

	double avgBLEUDifference = 0.0;
	vector<PROTranslationPair> tempTranslationPairs;

	int activeKBestListSizeMinusOne = activeKBestList.size() - 1;
	if (activeKBestList.size() > 1) {
		// Sample training pairs for the current sentence.
		for (int ii = 0; ii < gamma; ++ii) {
			// Get a random number between 0 and activeKBestList.size() - 1, inclusive.
			int randIndex1 = getRandomInteger(activeKBestListSizeMinusOne);
			// Get another one.
			int randIndex2 = getRandomInteger(activeKBestListSizeMinusOne);
			// Ensure that we get one that is different from randIndex1.
			while (randIndex2 == randIndex1) {
				randIndex2 = getRandomInteger(activeKBestListSizeMinusOne);
			}
			// Get the sentence-level BLEU scores of the two sampled translations.
			double randIndex1BLEU = activeKBestList[randIndex1].bleuScore;
			double randIndex2BLEU = activeKBestList[randIndex2].bleuScore;
			double sentenceBLEUDifference = randIndex1BLEU - randIndex2BLEU;
			double absSentenceBLEUDifference = abs(sentenceBLEUDifference);
			if (useSigmoid) {
				avgBLEUDifference += absSentenceBLEUDifference;

				PROTranslationPair newPair;
				newPair.index1 = randIndex1;
				newPair.index2 = randIndex2;
				newPair.bleuDifference = sentenceBLEUDifference;
				newPair.absBLEUDifference = absSentenceBLEUDifference;
				tempTranslationPairs.push_back(newPair);
			} else {
				// If the BLEU difference exceeds the alpha threshold or if we're not doing any subsampling, keep the pair.
				if (absSentenceBLEUDifference >= proAlphaThreshold || proNoSubsampling) {
					//	cerr << "Diff = " << sentenceBLEUDifference << ", adding pair" << endl;

					PROTranslationPair newPair;
					newPair.index1 = randIndex1;
					newPair.index2 = randIndex2;
					newPair.bleuDifference = sentenceBLEUDifference;
					newPair.absBLEUDifference = absSentenceBLEUDifference;
					translationPairs.push_back(newPair);
				//} else {
				//	cerr << "Diff = " << sentenceBLEUDifference << ", NOT adding pair" << endl;
				}
			}
		}

		if (useSigmoid) {
			// Compute the average difference in sentence-level BLEU.
			avgBLEUDifference /= tempTranslationPairs.size();

			for (int i = 0; i < tempTranslationPairs.size(); ++i) {
				double p = 1.0 / (1.0 + exp(-avgBLEUDifference - tempTranslationPairs[i].absBLEUDifference));
				//cerr << "avg_diff=" << avgBLEUDifference << "  gdiff=" << tempTranslationPairs[i].absBLEUDifference << "  p=" << p << endl;
				double randDouble = ((double) rand()) / ((double) RAND_MAX);
				if (randDouble < p) translationPairs.push_back(tempTranslationPairs[i]);
			}
		}

		// Sort translationPairs in decreasing order according to BLEU differences.
		sort(translationPairs.begin(), translationPairs.end(), m_comparePROTranslationPairsObject);

	} else {
		cerr << "Only one translation in k-best list. Skipping translation pair sampling." << endl;
	}
}

void PBModel::proAddTrainingPairs(const vector<TRANSLATION>& activeKBestList,
		const vector<PROTranslationPair>& translationPairs, int xi, bool proBothExamples) {

	int numToAdd = xi;
	if (numToAdd > translationPairs.size()) numToAdd = translationPairs.size();
	for (int ii = 0; ii < numToAdd; ++ii) {
		int index1 = translationPairs[ii].index1;
		int index2 = translationPairs[ii].index2;
		FEAT_HASH sparseFeatureVectorDifference;
		sparseFeatureVectorDifference.set_empty_key(99999);	// TODO: comment out if using sparsehash instead of densehash
		//computeFeatureMapDifference(activeKBestList[index1].featureValues, activeKBestList[index2].featureValues, featureVectorDifference);
		computeSparseFeatureMapDifference(activeKBestList[index1].featureValues, activeKBestList[index2].featureValues, sparseFeatureVectorDifference);
		map<int, double> featureVectorDifference;
		copySparseFeatureValues(sparseFeatureVectorDifference, featureVectorDifference);

		bool pairLabel = true;
		if (translationPairs[ii].bleuDifference < 0.0) pairLabel = false;
		m_binaryClassifier.addTrainingExample(featureVectorDifference, pairLabel);
		if (proBothExamples) {
			//map<int, double> example2FeatureVectorDifference;
			FEAT_HASH example2SparseFeatureVectorDifference;
			example2SparseFeatureVectorDifference.set_empty_key(99999);	// TODO: comment out if using sparsehash instead of densehash
			//computeFeatureMapDifference(activeKBestList[index2].featureValues, activeKBestList[index1].featureValues, example2FeatureVectorDifference);
			computeSparseFeatureMapDifference(activeKBestList[index2].featureValues, activeKBestList[index1].featureValues, example2SparseFeatureVectorDifference);
			map<int, double> example2FeatureVectorDifference;
			copySparseFeatureValues(example2SparseFeatureVectorDifference, example2FeatureVectorDifference);
			m_binaryClassifier.addTrainingExample(example2FeatureVectorDifference, !pairLabel);
		}
	}
}

void PBModel::imputeConcaveTerm(double costCoefficient) {
	bool findMaxTrue = true;
	bool costAugmentedTrue = true;
	bool costAugmentedFalse = false;
	bool useModelScoreForMaxTrue = true;

	m_imputedIndices.clear();
	m_imputedIndices.reserve(m_kbestLists.size());
	for (unsigned int i = 0; i < m_kbestLists.size(); ++i) {
		// Rescore the k-best list entries for when we call this on non-initial iterations of CCCP.
		rescoreKBestList(m_kbestLists[i], false);

		// We first find the index of the translation that maximizes the concave term for each loss function.
		int imputedIndex = -1;
		if (m_loss == RAMP1 || m_loss == HARDSOFTRAMP) {
			// For ramp and hardsoftramp, the concave term is simply a max over model scores.
			imputedIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedFalse, costCoefficient, useModelScoreForMaxTrue);
		} else if (m_loss == RAMP2 || m_loss == RAMP3 || m_loss == HARDSOFTRAMP_OCH || m_loss == HARDSOFTRAMP_SMM) {
			// For the other loss functions, the concave term is cost-diminished decoding, so we multiply the cost by -1.
			imputedIndex = findOptimalTranslation(m_kbestLists[i], findMaxTrue, costAugmentedTrue, -1.0 * costCoefficient, useModelScoreForMaxTrue);
		} else {
			cerr << "Error: unexpected loss function in imputeConcaveTerm(): " << m_loss << endl;
		}
		if (imputedIndex == -1) cerr << "Error: imputedIndex == -1! i = " << i << endl;
		m_imputedIndices.push_back(imputedIndex);
	}
}

double PBModel::computeCost(double costCoefficient, double bleuScore) {
	return costCoefficient - (costCoefficient * bleuScore);
}

// i is the index of the current k-best list we're looking at.
double PBModel::computeLossGradientForSentence(int i, double costCoefficient) {
	double lossValue = 0.0;
	FEAT_HASH::iterator iter;

	if (m_loss == ALLPAIRS_SOFTHINGE || m_loss == SOFTRAMP1 || m_loss == SOFTRAMP2 || m_loss == SOFTRAMP3) {
		double term1LogDenom = neg_inf;
		double term2LogDenom = neg_inf;

		vector<double> G1(G.size(), neg_inf);
		vector<double> G2 = G1;
		vector<bool> invertSigns(G.size(), false);

		// Go through every entry in the k-best list (this is the sum over <y,h> pairs).
		for (unsigned int j = 0; j < m_kbestLists[i].size(); ++j) {
			double term1DenomTerm = 0.0;
			double term2DenomTerm = 0.0;

			double costValue = computeCost(costCoefficient, m_kbestLists[i][j].bleuScore);

			if (m_loss == SOFTRAMP1) {
				// Term 1 requires summing the exp of the model scores.
				term1DenomTerm = m_kbestLists[i][j].score;
				// Term 2 requires summing the exp of the model score plus the cost.
				term2DenomTerm = m_kbestLists[i][j].score + costValue;
			} else if (m_loss == SOFTRAMP2) {
				// Term 1 requires summing the exp of the model score minus the cost.
				term1DenomTerm = m_kbestLists[i][j].score - costValue;
				// Term 2 requires summing the exp of the model scores.
				term2DenomTerm = (m_kbestLists[i][j].score);
			} else if (m_loss == SOFTRAMP3) {
				// Term 1 requires summing the exp of the model score minus the cost.
				term1DenomTerm = m_kbestLists[i][j].score - costValue;
				// Term 2 requires summing the exp of the model score plus the cost.
				term2DenomTerm = m_kbestLists[i][j].score + costValue;
			} else if (m_loss == ALLPAIRS_SOFTHINGE) {
				term1DenomTerm = (-m_kbestLists[i][j].score) - costValue;
				term2DenomTerm = m_kbestLists[i][j].score + costValue;
			}

			if (term1DenomTerm == neg_inf || term1DenomTerm != term1DenomTerm) {
				cerr << "Error: term1DenomTerm == " << term1DenomTerm << ", i = " << i << ", j = " << j << endl;
			}
			if (term2DenomTerm == neg_inf || term2DenomTerm != term2DenomTerm) {
				cerr << "Error: term2DenomTerm == " << term2DenomTerm << ", i = " << i << ", j = " << j << endl;
			}
			term1LogDenom = logAdd(term1LogDenom, term1DenomTerm);
			term2LogDenom = logAdd(term2LogDenom, term2DenomTerm);
			if (term1LogDenom == neg_inf || term1LogDenom != term1LogDenom) {
				cerr << "Error: term1LogDenom == " << term1LogDenom << ", i = " << i << ", j = " << j << endl;
			}
			if (term2LogDenom == neg_inf || term2LogDenom != term2LogDenom) {
				cerr << "Error: term2LogDenom == " << term2LogDenom << ", i = " << i << ", j = " << j << endl;
			}

			// Go through all nonzero features in the current entry on the current k-best list.
			// This is the same operation for all 4 loss functions, so we use the same code; there are just different ways
			// of computing each term in the denominator of terms 1 and 2 for each loss function.
			for (iter = m_kbestLists[i][j].featureValues.begin(); iter != m_kbestLists[i][j].featureValues.end(); ++iter) {
				double featureValue = iter->second;
				if (featureValue != 0.0) {
					if (featureValue < 0.0) {
						featureValue = -featureValue;
						invertSigns[iter->first] = true;
					}
					G1[iter->first] = logAdd(G1[iter->first], log(featureValue) + term1DenomTerm);
					G2[iter->first] = logAdd(G2[iter->first], log(featureValue) + term2DenomTerm);
				}
			}
		}

		// Go back through the gradient vectors and divide by the respective denominators, then add the two terms together.
		for (unsigned int j = 0; j < G.size(); ++j) {
			G1[j] -= term1LogDenom;
			G2[j] -= term2LogDenom;
			G1[j] = exp(G1[j]);
			G2[j] = exp(G2[j]);
			if (invertSigns[j]) {
				G1[j] = -G1[j];
				G2[j] = -G2[j];
			}
			G[j] += (G2[j] - G1[j]);
			// old subgradient line:
			//G[j] = G2[j] - G1[j];
			//	cerr << "before reg\t" << i << "\t" << j << "\tG for " << getParameterName(j) << " = " << G[j] << endl;
		}

		// Compute the loss for the current training example.
		lossValue += term2LogDenom;
		if (m_loss == ALLPAIRS_SOFTHINGE) {
			lossValue += term1LogDenom;
		} else {
			lossValue -= term1LogDenom;
		}
	} else if (m_loss == BAYESRISK) {
		double logNumer = neg_inf;
		double logDenom = neg_inf;

		vector<double> G1(G.size(), neg_inf);
		vector<double> G2 = G1;
		//vector<double> G3 = G1;
		vector<bool> invertSigns(G.size(), false);

		// Go through every entry in the k-best list (this is the sum over <y,h> pairs).
		for (unsigned int j = 0; j < m_kbestLists[i].size(); ++j) {
			// The current term in the denominator is the exp of the model score (we store it in the log domain).
			double denomTerm = m_kbestLists[i][j].score;
			double costValue = computeCost(costCoefficient, m_kbestLists[i][j].bleuScore);

			logDenom = logAdd(logDenom, denomTerm);
			if (logDenom == neg_inf || logDenom != logDenom) {
				cerr << "Error: logDenom == " << logDenom << ", i = " << i << ", j = " << j << endl;
			}
			logNumer = logAdd(logNumer, log(costValue) + denomTerm);
			if (logNumer != logNumer) {
				cerr << "Error: logNumer == " << logNumer << ", i = " << i << ", j = " << j << endl;
			}
			// Go through all nonzero features in the current entry on the current k-best list.
			for (iter = m_kbestLists[i][j].featureValues.begin(); iter != m_kbestLists[i][j].featureValues.end(); ++iter) {
				double featureValue = iter->second;
				if (featureValue != 0.0) {
					if (featureValue < 0.0) {
						featureValue = -featureValue;
						invertSigns[iter->first] = true;
					}
					G1[iter->first] = logAdd(G1[iter->first], log(featureValue) + log(costValue) + denomTerm);
					G2[iter->first] = logAdd(G2[iter->first], log(featureValue) + denomTerm);
					//G3[iter->first] = logAdd(G3[iter->first], log(costValue) + denomTerm);
				}
			}
		}
		double logExpectedCost = logNumer - logDenom;
		// Go back through the gradient vectors and divide by the denominator, then multiply the second and third terms together and add to the first.
		for (unsigned int j = 0; j < G.size(); ++j) {
			G1[j] -= logDenom;
			G2[j] -= logDenom;
			//G3[j] -= logDenom;
			// Log-multiply terms 2 and 3 (this is the product of expectations of the feature and the cost).
			G2[j] = G2[j] + logExpectedCost;
			G1[j] = exp(G1[j]);
			G2[j] = exp(G2[j]);
			if (invertSigns[j]) {
				G1[j] = -G1[j];
				G2[j] = -G2[j];
			}
			G[j] += (G1[j] - G2[j]);
		}

		// Add the loss for the current training example.
		lossValue += exp(logExpectedCost);
	} else if (m_loss == HARDSOFTRAMP || m_loss == HARDSOFTRAMP_OCH || m_loss == HARDSOFTRAMP_SMM) {
		double term2LogDenom = neg_inf;

		vector<double> G1(G.size(), 0.0);
		vector<double> G2(G.size(), neg_inf);
		vector<bool> invertSigns(G.size(), false);

		// Go through all nonzero features in the imputed term and add them to G1.
		int towardsIndex = m_imputedIndices[i];
		for (iter = m_kbestLists[i][towardsIndex].featureValues.begin(); iter != m_kbestLists[i][towardsIndex].featureValues.end(); ++iter) {
			double featureValue = iter->second;
			if (featureValue != 0.0) {
				G1[iter->first] = featureValue;
			}
		}

		// Go through every entry in the k-best list for sentence i (this is the sum over <y,h> pairs).
		for (unsigned int j = 0; j < m_kbestLists[i].size(); ++j) {
			double term2DenomTerm = 0.0;

			double costValue = computeCost(costCoefficient, m_kbestLists[i][j].bleuScore);

			if (m_loss == HARDSOFTRAMP || m_loss == HARDSOFTRAMP_SMM) {
				// Term 2 requires summing the exp of the model score plus the cost.
				term2DenomTerm = m_kbestLists[i][j].score + costValue;
			} else if (m_loss == HARDSOFTRAMP_OCH) {
				// Term 2 requires summing the exp of the model scores.
				term2DenomTerm = m_kbestLists[i][j].score;
			} else {
				cerr << "Error: unexpected loss function in hardsoftramp section of computeLossGradientForSentence(). m_loss = " << m_loss << endl;
			}

			if (term2DenomTerm == neg_inf || term2DenomTerm != term2DenomTerm) {
				cerr << "Error: term2DenomTerm == " << term2DenomTerm << ", i = " << i << ", j = " << j << endl;
			}
			term2LogDenom = logAdd(term2LogDenom, term2DenomTerm);
			if (term2LogDenom == neg_inf || term2LogDenom != term2LogDenom) {
				cerr << "Error: term2LogDenom == " << term2LogDenom << ", i = " << i << ", j = " << j << endl;
			}

			// Go through all nonzero features in the current entry on the current k-best list.
			// This is the same operation for all 4 loss functions, so we use the same code; there are just different ways
			// of computing each term in the denominator of terms 1 and 2 for each loss function.
			for (iter = m_kbestLists[i][j].featureValues.begin(); iter != m_kbestLists[i][j].featureValues.end(); ++iter) {
				double featureValue = iter->second;
				if (featureValue != 0.0) {
					if (featureValue < 0.0) {
						featureValue = -featureValue;
						invertSigns[iter->first] = true;
					}
					G2[iter->first] = logAdd(G2[iter->first], log(featureValue) + term2DenomTerm);
				}
			}
		}

		// Go back through the G2 vector and divide by the denominator, then add the two terms together.
		for (unsigned int j = 0; j < G.size(); ++j) {
			G2[j] -= term2LogDenom;
			G2[j] = exp(G2[j]);
			if (invertSigns[j]) {
				G2[j] = -G2[j];
			}
			G[j] += (G2[j] - G1[j]);
		}

		// Compute the loss for the current training example.
		// Subtract the score of the imputed output.
		lossValue -= m_kbestLists[i][towardsIndex].score;
		lossValue += term2LogDenom;
	} else if (m_loss == ALLPAIRS_EXP || m_loss == ALLPAIRS_EXPCOST) {
		// For all-pairs exp loss, we have to go through all *pairs* of outputs on the current k-best list.

		double logLoss = neg_inf;

		vector<double> G1(G.size(), neg_inf);
		vector<double> G2 = G1;
		vector<bool> invertSigns(G.size(), false);

		// Go through every entry in the k-best list.
		for (unsigned int j1 = 0; j1 < m_kbestLists[i].size(); ++j1) {
			double cost1 = computeCost(costCoefficient, m_kbestLists[i][j1].bleuScore);
			// Go through every entry after j1 to ensure that we don't see any pairs more than once.
			for (unsigned int j2 = j1 + 1; j2 < m_kbestLists[i].size(); ++j2) {
				double cost2 = computeCost(costCoefficient, m_kbestLists[i][j2].bleuScore);
				if (cost1 != cost2) {
					// Figure out which translation has higher cost.
					double higherCost = cost2;
					double lowerCost = cost1;
					int lowerCostIndex = j1;
					int higherCostIndex = j2;
					if (cost1 > cost2) {
						lowerCostIndex = j2;
						higherCostIndex = j1;
						higherCost = cost1;
						lowerCost = cost2;
					}
					double score1 = m_kbestLists[i][higherCostIndex].score;
					double score2 = m_kbestLists[i][lowerCostIndex].score;
					if (score1 == 0.0 || score1 == neg_inf || score1 != score1) {
						cerr << "Error: score1 = " << score1 << ", i = " << i << ", higherCostIndex = " << higherCostIndex << ", j1 = " << j1 << ", j2 = " << j2 << endl;
					}
					if (score2 == 0.0 || score2 == neg_inf || score2 != score2) {
						cerr << "Error: score2 = " << score2 << ", i = " << i << ", lowerCostIndex = " << lowerCostIndex << ", j1 = " << j1 << ", j2 = " << j2 << endl;
					}
					// Compute the log of the loss for the current pair. The loss is large when the translation with
					// higher cost has a much higher score than the translation with lower cost.
					double scoreDifference = m_kbestLists[i][higherCostIndex].score - m_kbestLists[i][lowerCostIndex].score;
					double costDifference = higherCost - lowerCost;

					if (scoreDifference == neg_inf || scoreDifference != scoreDifference) {
						cerr << "Error: scoreDifference = " << scoreDifference << ", i = " << i << ", j1 = " << j1 << ", j2 = " << j2 << endl;
					}
					double currTerm = scoreDifference;
					if (m_loss == ALLPAIRS_EXPCOST) {
						currTerm += costDifference;
					}
					// Go through all nonzero features in the higher-cost output and add the feature
					// value (times the score difference) to G1.
					for (iter = m_kbestLists[i][higherCostIndex].featureValues.begin(); iter != m_kbestLists[i][higherCostIndex].featureValues.end(); ++iter) {
						double featureValue = iter->second;
						if (featureValue != 0.0) {
							if (featureValue < 0.0) {
								featureValue = -featureValue;
								invertSigns[iter->first] = true;
							}
							G1[iter->first] = logAdd(G1[iter->first], log(featureValue) + currTerm);
							if (G1[iter->first] == neg_inf || G1[iter->first] != G1[iter->first]) {
								cerr << "Error: G1[" << iter->first << "]== " << G1[iter->first] << ", i = " << i << ", j1 = " << j1 << ", j2 = " << j2 << endl;
							}
						}
					}

					// Go through all nonzero features in the lower-cost output and add the feature
					// value (times the score difference) to G2.
					for (iter = m_kbestLists[i][lowerCostIndex].featureValues.begin(); iter != m_kbestLists[i][lowerCostIndex].featureValues.end(); ++iter) {
						double featureValue = iter->second;
						if (featureValue != 0.0) {
							if (featureValue < 0.0) {
								featureValue = -featureValue;
								invertSigns[iter->first] = true;
							}
							G2[iter->first] = logAdd(G2[iter->first], log(featureValue) + currTerm);
							if (G2[iter->first] == neg_inf || G2[iter->first] != G2[iter->first]) {
								cerr << "Error: G2[" << iter->first << "]== " << G2[iter->first] << ", i = " << i << ", j1 = " << j1 << ", j2 = " << j2 << endl;
							}
						}
					}

					logLoss = logAdd(logLoss, currTerm);
					if (logLoss == neg_inf || logLoss != logLoss) {
						cerr << "Error: logLoss == " << logLoss << ", i = " << i << ", j1 = " << j1 << ", j2 = " << j2 << endl;
					}
				}
			}
		}

		// Go back through the gradient vectors and divide by the respective denominators, then add the two terms together.
		for (unsigned int j = 0; j < G.size(); ++j) {
			G1[j] = exp(G1[j]);
			G2[j] = exp(G2[j]);
			if (invertSigns[j]) {
				G1[j] = -G1[j];
				G2[j] = -G2[j];
			}
			G[j] += (G1[j] - G2[j]);
			if (G[j] != G[j]) {
				cerr << "Error: param = " << j << " (" << getParameterName(j) << "), G1 = " << G1[j] << ", G2 = " << G2[j] << endl;
			}

		}

		// Add the loss for the current training example.
		lossValue += exp(logLoss);


	} else if (m_loss == LINEAR_REGRESSION) {
		// Go through every entry in the k-best list (this is the sum over <y,h> pairs).
		for (unsigned int j = 0; j < m_kbestLists[i].size(); ++j) {
			// The current term in the denominator is the exp of the model score (we store it in the log domain).
			double scoreValue = m_kbestLists[i][j].score;

			double gainValue = costCoefficient * m_kbestLists[i][j].bleuScore;
			double scoreGainDiff = scoreValue - gainValue;
			lossValue += (scoreGainDiff * scoreGainDiff);

			// Go through all nonzero features in the current entry on the current k-best list.
			for (iter = m_kbestLists[i][j].featureValues.begin(); iter != m_kbestLists[i][j].featureValues.end(); ++iter) {
				double featureValue = iter->second;
				if (featureValue != 0.0) {
					G[iter->first] += (2.0 * scoreGainDiff * featureValue);
				}
			}
		}
	}
	return lossValue;
}

// Compute sentence-level BLEU scores for all translations on the k-best list.
// Returns index of translation with highest BLEU score.
int PBModel::computeSentenceLevelBLEUScores(vector<TRANSLATION>& kBestList,
	const vector< vector<string> >& currLineMultiReferenceLines) {
	// Create an mteval object for the reference sentences for the current line.
	MTEval mteval(currLineMultiReferenceLines, m_costMetric);
	int bestBLEUIndex = 0;
	// Compute sentence-level BLEU scores for all translations on the k-best list.
	for (unsigned int ii = 0; ii < kBestList.size(); ++ii) {
		// Create a dummy BLEU sufficient statistics variable that will get filled in when sentence BLEU is computed.
		BLEUSuffStats bleuSuffStats;
		kBestList[ii].bleuScore = mteval.computeSentenceBLEU(kBestList[ii].text, bleuSuffStats);
		if (ii > 0 && kBestList[ii].bleuScore > kBestList[bestBLEUIndex].bleuScore) {
			bestBLEUIndex = ii;
		}
	}
	return bestBLEUIndex;
}

// Compute sentence-level BLEU scores for all translations on the k-best list. Uses provided MTEval object.
// Returns index of translation with highest BLEU score.
int PBModel::computeSentenceLevelBLEUScores(vector<TRANSLATION>& kBestList, MTEval& mteval, unsigned int kbestListStartingIndex) {
	int bestBLEUIndex = 0;
	// Compute sentence-level BLEU scores for all translations on the k-best list.
	for (unsigned int ii = 0; ii < kBestList.size(); ++ii) {
		if (ii >= kbestListStartingIndex) {
			// Create a dummy BLEU sufficient statistics variable that will get filled in when sentence BLEU is computed.
			BLEUSuffStats bleuSuffStats;
			kBestList[ii].bleuScore = mteval.computeSentenceBLEU(kBestList[ii].text, bleuSuffStats);
		}
		if (ii > 0) {
			if (kBestList[ii].bleuScore > kBestList[bestBLEUIndex].bleuScore) {
				bestBLEUIndex = ii;
			}
		}
	}
	return bestBLEUIndex;
}


// Fills subset with the subsequences of entries (across ALL lists in full's first dimension) from start to end in full's second dimension.
// That is, if full is k x m, start = 0, and end = n < m, subset will be k x (n+1). Each of the k vectors in subset
// will contain entries from start to end in each of the k vectors in full.
void PBModel::getReferenceLinesSubset(vector< vector<string> >& full, unsigned int start, unsigned int end, vector< vector<string> >& subset) {
	subset.clear();
	subset.reserve(end - start + 1);
	unsigned int numDimensions = full.size();
	if (numDimensions > 0) {
		// Create vectors for each of the dimensions in full.
		for (unsigned int i = 0; i < numDimensions; ++i) {
			vector<string> curr;
			subset.push_back(curr);
			// Check the bounds of end.
			if (end >= full[i].size()) {
				end = full[i].size() - 1;
			}
		}
		if (start < 0) start = 0;
		for (unsigned int i = start; i <= end; ++i) {
			for (unsigned int j = 0; j < numDimensions; ++j) {
				subset[j].push_back(full[j][i]);
			}
		}
	}
}

// This comparison function will cause the translation with the highest BLEU score
// to be first in a vector after a call to sort().
bool compareTranslationBLEUScores(const pair<double, BLEUSuffStats> &a, const pair<double, BLEUSuffStats> &b) {
	return a.first > b.first;
}


